# coding=utf-8
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

"""
Function:
Shape Format Conversion class. This class mainly involves the convert_shape function.
"""

import argparse
import os
import sys
import re

import numpy as np

from msprobe.msaccucmp.cmp_utils import utils, utils_type, path_check
from msprobe.msaccucmp.cmp_utils import common
from msprobe.msaccucmp.cmp_utils import log
from msprobe.msaccucmp.cmp_utils.utils import safe_path_string
from msprobe.msaccucmp.cmp_utils.constant.const_manager import ConstManager, DD
from msprobe.msaccucmp.cmp_utils.multi_process.multi_convert_process import MultiConvertProcess
from msprobe.msaccucmp.cmp_utils.reg_manager import RegManager
from msprobe.msaccucmp.cmp_utils.file_utils import FileUtils
from msprobe.msaccucmp.cmp_utils.constant.compare_error import CompareError
from msprobe.msaccucmp.format_manager.format_manager import FormatManager
from msprobe.msaccucmp.format_manager.format_manager import SrcToDest
from msprobe.msaccucmp.format_manager.format_manager import ShapeConversion
from msprobe.msaccucmp.conversion.tensor_conversion import TensorConversion
from msprobe.msaccucmp.dump_parse import dump, dump_utils, mapping
from msprobe.msaccucmp.dump_parse.dump_data_object import DumpDataObj, DumpTensor


def _check_shape_valid(shape_str: str) -> list:
    shape = []
    if shape_str:
        # check shape is valid
        shape_pattern = re.compile(RegManager.SUPPORT_SHAPE_PATTERN)
        match = shape_pattern.match(shape_str)
        if match is None:
            log.print_error_log('The shape(%r) is invalid. The supported formats are "1,3,224,224".' % shape_str)
            raise CompareError(CompareError.MSACCUCMP_INVALID_SHAPE_ERROR)
        split_list = shape_str.split(',')

        for dim_str in split_list:
            dim = int(dim_str)
            if dim > 0:
                shape.append(dim)
            else:
                log.print_error_log('The shape(%r) is invalid. Each dimension must be greater than 0.' % shape_str)
                raise CompareError(CompareError.MSACCUCMP_INVALID_SHAPE_ERROR)
    return shape


class ShapeConversionMain:
    """
    The class for format conversion
    """

    def __init__(self: any) -> None:
        parse = argparse.ArgumentParser()
        parse.add_argument("-i", dest="dump_file_path", default="", type=safe_path_string,
                           help="<Required> the dump file path",
                           required=True)
        parse.add_argument("-format", dest="format", default="",
                           help="<Required> the format to transfer",
                           required=True)
        parse.add_argument("-o", dest="output_path", default="", type=safe_path_string,
                           help="<Required> the output path", required=True)
        parse.add_argument('-tensor', dest="tensor", default="output",
                           help="<Optional> the tensor, input or output")
        parse.add_argument('-index', dest="index", default="0",
                           help="<Optional> the index for tensor")
        parse.add_argument(
            '-shape', dest="shape",
            help="<Optional> the shape for format transfer, currently only"
                 " used for FRACTAL_NZ conversion, shape format is "
                 "([0-9]+,)+[0-9]+, such as 1,3,224,224")
        parse.add_argument("-custom", dest="custom_path", default="", type=safe_path_string,
                           help="<Optional> user-defined path, including format conversion",
                           required=False)
        args, _ = parse.parse_known_args(sys.argv[1:])
        self.dump_file_path = os.path.realpath(args.dump_file_path)
        if os.path.islink(os.path.abspath(args.output_path)):
            log.print_error_log('The path "%r" is a softlink, not permitted.' % args.output_path)
            raise CompareError(CompareError.MSACCUCMP_INVALID_PATH_ERROR)
        self.output_path = os.path.realpath(args.output_path)
        format_str = args.format
        shape_str = args.shape
        self.manager = FormatManager(args.custom_path)
        index_str = args.index
        self.tensor = args.tensor
        if not RegManager.match_pattern(RegManager.NUMBER_PATTERN, index_str):
            log.print_only_support_error('tensor index', index_str, 'natural number')
            raise CompareError(CompareError.MSACCUCMP_INVALID_PARAM_ERROR)
        self.index = int(index_str)
        # check format is valid
        if format_str not in ConstManager.STRING_TO_FORMAT_MAP:
            log.print_error_log('The format "%r" is not supported. '
                                'Please check the format.' % format_str)
            raise CompareError(CompareError.MSACCUCMP_INVALID_PARAM_ERROR)
        self.format_to = ConstManager.STRING_TO_FORMAT_MAP.get(format_str)
        self.shape = _check_shape_valid(shape_str)

    def process(self: any) -> int:
        """
        Check arguments valid, then convert shape and slice data
        :return: error code
        """
        ret = self.check_arguments_valid()
        if ret != CompareError.MSACCUCMP_NONE_ERROR:
            return ret
        try:
            dump_data = dump_utils.parse_dump_file(self.dump_file_path, ConstManager.OLD_DUMP_TYPE)
        except CompareError as error:
            return error.code

        if self.tensor == ConstManager.INPUT:
            tensor_list = dump_data.input_data
        else:
            tensor_list = dump_data.output_data
        if self.index >= len(tensor_list):
            log.print_out_of_range_error('', self.tensor, self.index, '[0, %d)' % len(tensor_list))
            return CompareError.MSACCUCMP_INDEX_OUT_OF_BOUNDS_ERROR

        try:
            self._convert_format(tensor_list[self.index], self.index)
        except CompareError as error:
            return error.code
        return ret

    def check_arguments_valid(self: any) -> int:
        """
        Check arguments valid
        :return error code
        """
        self.manager.check_arguments_valid()

        # check tensor valid
        if self.tensor not in ConstManager.SUPPORT_DETAIL_TYPE:
            log.print_only_support_error('tensor', self.tensor, ConstManager.SUPPORT_DETAIL_TYPE)
            return CompareError.MSACCUCMP_INVALID_PARAM_ERROR

        # check dump file path is valid
        ret = path_check.check_path_valid(
            self.dump_file_path, True, False, path_check.PathType.File)
        if ret != CompareError.MSACCUCMP_NONE_ERROR:
            return ret

        # check output path is valid
        ret = path_check.check_output_path_valid(self.output_path, True)
        if ret != CompareError.MSACCUCMP_NONE_ERROR:
            return ret

        log.print_info_log('Start to transfer format for the dump file "%r".' % self.dump_file_path)
        log.print_info_log('The target format is %s for %r:%d.'
                           % (common.get_format_string(self.format_to), self.tensor, self.index))
        return CompareError.MSACCUCMP_NONE_ERROR

    def _convert_format(self: any, tensor: any, index: int) -> None:
        if tensor.tensor_format == self.format_to:
            log.print_info_log("There is no need to transfer format because the format (%s) is the same."
                               % common.get_format_string(tensor.tensor_format))
            return
        # deserialize dump data
        dump_data_array = tensor.data

        real_format = tensor.tensor_format
        # check the 4-dimension valid
        if tensor.tensor_format == DD.FORMAT_HWCN or tensor.tensor_format == DD.FORMAT_NCHW \
                or tensor.tensor_format == DD.FORMAT_NHWC:
            if len(tensor.shape) != ConstManager.FOUR_DIMS_LENGTH:
                log.print_warn_log(
                    'The format(%s) of the dump data is 4 dimensions, but the shape%s is not 4 dimensions but ND.'
                    % (common.get_format_string(tensor.tensor_format), utils.convert_shape_to_string(tensor.shape)))
                real_format = DD.FORMAT_ND
        log.print_info_log("Before transferring the format, the dump data is %s%s."
                           % (common.get_format_string(real_format),
                              utils.convert_shape_to_string(tensor.shape)))
        if real_format == DD.FORMAT_FRACTAL_NZ:
            utils.check_shape_valid_in_nz(self.shape, tensor.shape)

        dump_data_np = self.manager.execute_format_convert(
            SrcToDest(real_format, self.format_to, tensor.shape, self.shape),
            dump_data_array, {'group': common.get_sub_format(tensor)})
        log.print_info_log("After transferring the format, the dump data is %s%s."
                           % (common.get_format_string(self.format_to),
                              utils.convert_shape_to_string(dump_data_np.shape)))
        # save numpy data to file
        output_file_path = os.path.join(self.output_path, '%s.%s.%d.%s.npy'
                                        % (os.path.basename(self.dump_file_path), self.tensor, index,
                                           utils.get_string_from_list(dump_data_np.shape, 'x')))
        np.save(output_file_path, dump_data_np)
        log.print_info_log('The dump data for %r:%d has been saved to "%r".'
                           % (self.tensor, index, output_file_path))


class FormatConversionMain:
    """
    The class for format conversion
    """

    def __init__(self: any, arguments: any = None) -> None:
        self.path_str = arguments.dump_path
        if os.path.islink(os.path.abspath(arguments.output_path)):
            log.print_error_log('The path "%r" is a softlink, not permitted.' % arguments.output_path)
            raise CompareError(CompareError.MSACCUCMP_INVALID_PATH_ERROR)
        self.output_path = os.path.realpath(arguments.output_path)
        self.attr = {
            'dump_version': arguments.dump_version,
            'output_file_type': arguments.output_file_type
        }
        custom_script_path = arguments.custom_script_path if arguments.custom_script_path else ''
        self.manager = FormatManager(custom_script_path)
        self.one_file_info = None
        if ',' not in self.path_str and os.path.isfile(os.path.realpath(self.path_str)):
            tensor_type = ''
            index_str = '0'
            if arguments.input:
                tensor_type = ConstManager.INPUT
                index_str = arguments.input
            if arguments.output:
                tensor_type = ConstManager.OUTPUT
                index_str = arguments.output
            if not RegManager.match_pattern(RegManager.NUMBER_PATTERN, index_str):
                log.print_only_support_error('tensor index', index_str, 'natural number')
                raise CompareError(CompareError.MSACCUCMP_INVALID_PARAM_ERROR)
            self.index = int(index_str)
            shape = _check_shape_valid(arguments.shape)
            self.one_file_info = {'tensor': tensor_type, 'index': int(index_str), 'shape': shape}

        format_str = arguments.format
        # check format is valid
        if format_str not in ConstManager.STRING_TO_FORMAT_MAP:
            log.print_error_log('The format "%s" is not supported. Please check the format.' % format_str)
            raise CompareError(CompareError.MSACCUCMP_INVALID_PARAM_ERROR)
        self.format_to = ConstManager.STRING_TO_FORMAT_MAP.get(format_str)
        self.multi_process = None
        self.input_path = []

    def convert_format(self: any) -> int:
        """
        Convert dump data to numpy or bin file
        :return error_code
        """
        # 1. check arguments valid
        ret = self.check_arguments_valid()
        if ret != CompareError.MSACCUCMP_NONE_ERROR:
            log.print_error_log('Check arguments Failed. Please recheck the arguments')
        # 2. convert format for dump data
        if len(self.input_path) == 1 and os.path.isfile(self.input_path[0]):
            ret, _ = self._convert_format_for_one_file(self.input_path[0])
            if ret != CompareError.MSACCUCMP_NONE_ERROR:
                log.print_error_log('Failed to convert format for %s".' % self.input_path[0])
            return ret
        return self.multi_process.process()

    def check_arguments_valid(self: any) -> int:
        """
        Check arguments valid
        :return: error code
        """
        self.manager.check_arguments_valid()

        # check dump file path is valid
        self.input_path = path_check.get_path_list_for_str(self.path_str)
        self.multi_process = MultiConvertProcess(self._convert_format_for_one_file, self.input_path,
                                                 self.output_path)

        # check output path is valid
        ret = path_check.check_output_path_valid(self.output_path, True)
        if ret != CompareError.MSACCUCMP_NONE_ERROR:
            return ret

        log.print_info_log('The target format is "%s".'
                           % common.get_format_string(self.format_to))

        return CompareError.MSACCUCMP_NONE_ERROR

    def _save_to_file(self: any, *args: any) -> None:
        output_format, dump_data_np, tensor_type, index, dump_file_path, tensor, op_name = args
        if self.attr.get('output_file_type') == 'npy':
            file_name = '%s.%s.%d.%s.npy' % (os.path.basename(dump_file_path), tensor_type, index,
                                             utils.get_string_from_list(dump_data_np.shape, 'x'))
            file_name = FileUtils.handle_too_long_file_name(
                file_name, '.npy', os.path.join(self.output_path, ConstManager.MAPPING_FILE_NAME))
        elif self.attr.get('output_file_type') == 'msnpy':
            file_name = utils.make_msnpy_file_name(dump_file_path, op_name, tensor_type, index, output_format)
            file_name = FileUtils.handle_too_long_file_name(
                file_name, '.npy', os.path.join(self.output_path, ConstManager.MAPPING_FILE_NAME))
        else:
            file_name = '%s.%s.%d.%s.%s.%s.bin' % (os.path.basename(dump_file_path), tensor_type, index,
                                                   utils.get_string_from_list(dump_data_np.shape, '_'),
                                                   np.dtype(common.get_dtype_by_data_type(tensor.data_type)).name,
                                                   common.get_format_string(output_format))
            file_name = FileUtils.handle_too_long_file_name(
                file_name, '.bin', os.path.join(self.output_path, ConstManager.MAPPING_FILE_NAME))
        output_file_path = os.path.join(self.output_path, file_name)
        FileUtils.save_array_to_file(output_file_path, dump_data_np,
                                     self.attr.get('output_file_type') != 'bin', dump_data_np.shape)
        log.print_info_log(
            'The data of %s:%s has been saved to "%r".' % (tensor_type, index, output_file_path))

    def _get_format_and_shape(self: any, tensor: any, index: int, tensor_type: str,
                              dump_file_path: str):
        # check the 4-dimension valid
        real_format = tensor.tensor_format
        if tensor.tensor_format == DD.FORMAT_HWCN or tensor.tensor_format == DD.FORMAT_NCHW \
                or tensor.tensor_format == DD.FORMAT_NHWC:
            if len(tensor.shape) != ConstManager.FOUR_DIMS_LENGTH:
                log.print_warn_log(
                    'The format(%s) of the dump data is 4 dimensions, but the shape %s is not 4 dimensions,'
                    ' the real format is ND for %s:%d of "%r".'
                    % (common.get_format_string(tensor.tensor_format), utils.convert_shape_to_string(tensor.shape),
                       tensor_type, index, dump_file_path))
                real_format = DD.FORMAT_ND

        shape = []
        if self.one_file_info:
            shape = self.one_file_info.get('shape')
        if len(shape) == 0:
            shape = tensor.original_shape
        if real_format == DD.FORMAT_FRACTAL_NZ:
            utils.check_shape_valid_in_nz(shape, tensor.shape)

        return real_format, shape

    def _convert_format_for_one_tensor(self: any, *args: any) -> None:
        tensor, index, tensor_type, dump_file_path, op_name = args
        # deserialize dump data
        dump_data_array = tensor.data

        real_format, shape_to = self._get_format_and_shape(tensor, index, tensor_type, dump_file_path)

        log.print_info_log(
            "Before transferring the format, the data of %s:%d is %s%s."
            % (tensor_type, index, common.get_format_string(real_format),
               utils.convert_shape_to_string(tensor.shape)))
        src_to_dest = SrcToDest(real_format, self.format_to, tensor.shape, shape_to)
        if real_format == self.format_to:
            log.print_info_log(
                'There is no need to transfer format because the format (%s) '
                'is the same for %s:%d of "%r".'
                % (common.get_format_string(tensor.tensor_format), tensor_type,
                   index, dump_file_path))
            dump_data_np = ShapeConversion(self.manager).reshape(src_to_dest, dump_data_array)
        else:
            # convert shape
            dump_data_np = self.manager.execute_format_convert(
                src_to_dest, dump_data_array, {'group': common.get_sub_format(tensor)})
            log.print_info_log(
                "After transferring the format, the data of %s:%d is %s%s."
                % (tensor_type, index, common.get_format_string(self.format_to),
                   utils.convert_shape_to_string(dump_data_np.shape)))
        output_format = self.format_to
        # slice data
        try:
            if len(shape_to) > 0 and real_format != self.format_to:
                old_shape_str = utils.convert_shape_to_string(dump_data_np.shape)
                tensor_conversion = TensorConversion(None, self.manager, is_detail=False)
                dump_data_np = tensor_conversion.slice_data(dump_data_np, shape_to)
                log.print_info_log('The data of %s:%d has been sliced from %s to %s.' %
                                   (tensor_type, index, old_shape_str,
                                    utils.convert_shape_to_string(shape_to)))
        except (OSError, SystemError, ValueError, TypeError, RuntimeError,
                MemoryError, CompareError):
            log.print_error_log('Failed to slice data for %r from %s to %s.'
                                % (dump_file_path,
                                   utils.convert_shape_to_string(dump_data_np.shape),
                                   utils.convert_shape_to_string(shape_to)))
            output_format = tensor.tensor_format
        finally:
            pass
        # save numpy data to file
        self._save_to_file(output_format, dump_data_np, tensor_type, index, dump_file_path, tensor, op_name)

    def _save_file_for_convert_failed_tensor(self: any, *args: any) -> None:
        tensor, dump_path, tensor_type, index, op_name = args
        if self.attr.get('output_file_type') == 'msnpy':
            file_name = utils.make_msnpy_file_name(dump_path, op_name, tensor_type, index, tensor.tensor_format)
            file_name = FileUtils.handle_too_long_file_name(
                file_name, '.npy', os.path.join(self.output_path, ConstManager.MAPPING_FILE_NAME))
            output_file_path = os.path.join(self.output_path, file_name)
            FileUtils.save_array_to_file(output_file_path, tensor.data,
                                         self.attr.get('output_file_type') != 'bin',
                                         tensor.shape)
            log.print_info_log('The data of %s:%d has been parsed into "%r".'
                               % (tensor_type, index, output_file_path))

    def _convert_format_for_tensor(self: any, tensor_list: list, dump_file_path: str, tensor_type: str,
                                   op_name: str) -> (int, str):
        ret = CompareError.MSACCUCMP_NONE_ERROR
        msg = ""
        for index, tensor in enumerate(tensor_list):
            try:
                self._convert_format_for_one_tensor(tensor, index, tensor_type, dump_file_path, op_name)
            except (OSError, SystemError, ValueError, TypeError, RuntimeError,
                    MemoryError) as error:
                log.print_error_log('Failed to convert format for %s:%d of "%r". %s'
                                    % (tensor_type, index, dump_file_path, error))
                ret = CompareError.MSACCUCMP_UNKNOWN_ERROR
                msg += ',%s:%d' % (tensor_type, index)
                self._save_file_for_convert_failed_tensor(tensor, dump_file_path, tensor_type, index, op_name)
            except CompareError as error:
                log.print_error_log('Failed to convert format for %s:%d of "%r"'
                                    % (tensor_type, index, dump_file_path))
                ret = error.code
                msg += ',%s:%d' % (tensor_type, index)
                self._save_file_for_convert_failed_tensor(tensor, dump_file_path, tensor_type, index, op_name)
            finally:
                pass
        return ret, msg

    def _convert_format_for_input_or_output(self: any, dump_data: DumpDataObj, dump_file_path: str) -> None:
        if self.one_file_info.get('tensor') == ConstManager.INPUT:
            tensor_list = dump_data.input_data
        else:
            tensor_list = dump_data.output_data
        if self.one_file_info.get('index') >= len(tensor_list):
            log.print_error_log(
                'The %s index (%d) is out of range [0, %d). Please check the index.'
                % (self.one_file_info.get('tensor'), self.one_file_info.get('index'),
                   len(tensor_list)))
            raise CompareError(CompareError.MSACCUCMP_INDEX_OUT_OF_BOUNDS_ERROR)
        self._convert_format_for_one_tensor(
            tensor_list[self.one_file_info.get('index')], self.one_file_info.get('index'),
            self.one_file_info.get('tensor'), dump_file_path, dump_data.op_name)

    def _convert_format_for_tensor_list(self: any, dump_data: DumpDataObj, dump_file_path: str) -> (int, str):
        ret = CompareError.MSACCUCMP_NONE_ERROR
        msg = dump_file_path
        ret_input, msg_input = self._convert_format_for_tensor(dump_data.input_data, dump_file_path, ConstManager.INPUT,
                                                               dump_data.op_name)
        if ret_input != CompareError.MSACCUCMP_NONE_ERROR:
            ret = ret_input
        msg += msg_input
        ret_output, msg_output = self._convert_format_for_tensor(dump_data.output_data, dump_file_path,
                                                                 ConstManager.OUTPUT, dump_data.op_name)
        if ret_output != CompareError.MSACCUCMP_NONE_ERROR:
            ret = ret_output
        msg += msg_output
        return ret, msg

    def _convert_format_for_one_file(self: any, dump_file_path: str) -> (int, str):
        ret = CompareError.MSACCUCMP_NONE_ERROR
        msg = dump_file_path
        log.print_info_log('Start to transfer format for the dump file "%r".' % dump_file_path)
        try:
            dump_data = dump_utils.parse_dump_file(dump_file_path, self.attr.get('dump_version'))

        except CompareError as error:
            return error.code, msg

        try:
            if self.one_file_info and self.one_file_info.get('tensor'):
                self._convert_format_for_input_or_output(dump_data, dump_file_path)
            else:
                ret, msg = self._convert_format_for_tensor_list(dump_data, dump_file_path)
        except (OSError, SystemError, ValueError, TypeError, RuntimeError, MemoryError) as error:
            log.print_error_log('Failed to convert format for "%r". %s' % (dump_file_path, error))
            ret = CompareError.MSACCUCMP_UNKNOWN_ERROR
        except CompareError as error:
            ret = error.code
        return ret, msg



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

from abc import ABC, abstractmethod
import numpy as np
from msprobe.pytorch.api_accuracy_checker.compare.compare_utils import convert_str_to_float
from msprobe.pytorch.api_accuracy_checker.compare.algorithm import get_abs_bench_with_eps, get_abs_err, \
    get_finite_and_infinite_mask, get_small_value_mask
from msprobe.pytorch.api_accuracy_checker.precision_standard.standard_config import StandardConfig


class BaseCompare(ABC):
    """
    Base comparison class for benchmarking and device output.

    This class provides a foundation for comparing benchmark outputs with device outputs.
    It encapsulates the common logic for calculating accuracy metrics and
    provides a framework for subclasses to implement specific comparison logic.

    Attributes:
        bench_output (np.ndarray): The output from the benchmark.
        device_output (np.ndarray): The output from the device.
        compare_column (object): The column object to store comparison results.
        dtype (torch.dtype): The data type of the outputs.

    Methods:
        get_small_value_threshold(): Retrieves the small value threshold for the given data type.
        stat_abs_bench_with_eps(): Calculates the absolute benchmark output with epsilon.
        stat_abs_error(): Calculates the absolute error between the benchmark and device outputs.
        stat_finite_and_infinite_mask(): Generates masks for finite and infinite/NaN values.
        stat_small_value_mask(abs_bench, both_finite_mask, small_value): Creates a mask for small values.
        compare(): Performs the comparison and computes metrics.
        _pre_compare(): Pre-comparison hook for subclass-specific initialization.
        _compute_metrics(): Computes the comparison metrics.
        _post_compare(metrics): Post-comparison hook to update comparison results.

    Note:
        This class assumes that the input data is an instance of InputData containing the benchmark output,
        device output, comparison column, and data type. Subclasses should implement the _pre_compare,
        _compute_metrics, and _post_compare methods to provide specific comparison logic.

    See Also:
        InputData: The class containing input data for comparison.
        StandardConfig: The class containing standard configuration values.
    """
    def __init__(self, input_data):
        self.bench_output = input_data.bench_output
        self.device_output = input_data.device_output
        self.compare_column = input_data.compare_column
        self.dtype = input_data.dtype
        self.compare_algorithm = None

    @staticmethod
    def stat_small_value_mask(abs_bench, both_finite_mask, small_value):
        small_value_mask = get_small_value_mask(abs_bench, both_finite_mask, small_value)
        return small_value_mask
    
    @staticmethod
    def _get_rel_err(abs_err, abs_bench_with_eps):
        rel_err = abs_err / abs_bench_with_eps
        return rel_err
    
    @staticmethod
    def _get_normal_value_mask(both_finite_mask, small_value_mask):
        return np.logical_and(both_finite_mask, np.logical_not(small_value_mask))

    @abstractmethod
    def _pre_compare(self):
        raise NotImplementedError

    def get_small_value_threshold(self):
        small_value = StandardConfig.get_small_value(self.dtype, self.compare_algorithm)
        small_value_atol = StandardConfig.get_small_value_atol(self.dtype, self.compare_algorithm)
        return small_value, small_value_atol
    
    def stat_abs_bench_with_eps(self):
        abs_bench, abs_bench_with_eps = get_abs_bench_with_eps(self.bench_output, self.dtype)
        return abs_bench, abs_bench_with_eps
    
    def stat_abs_error(self):
        abs_err = get_abs_err(self.bench_output, self.device_output)
        return abs_err
    
    def stat_finite_and_infinite_mask(self):
        both_finite_mask, inf_nan_mask = get_finite_and_infinite_mask(self.bench_output, self.device_output)
        return both_finite_mask, inf_nan_mask

    def compare(self):
        self._pre_compare()
        metrics = self._compute_metrics()
        self._post_compare(metrics)

    def _compute_metrics(self):
        return {}
    
    def _post_compare(self, metrics):
        self.compare_column.update(metrics)


class BasePrecisionCompare:
    def __init__(self, input_data):
        self.row_npu = input_data.row_npu
        self.row_gpu = input_data.row_gpu
        self.dtype = input_data.dtype
        self.compare_column = input_data.compare_column
        self.compare_algorithm = None
    
    @staticmethod
    def _get_and_convert_value(row_data, column_name, device_type):
        value = row_data.get(column_name)
        if value is None:
            raise ValueError(f"{device_type} value for column '{column_name}' is None.")
        return convert_str_to_float(value)

    @abstractmethod
    def _get_status(self, metrics, inf_nan_consistency):
        pass

    @abstractmethod
    def _compute_ratio(self):
        pass

    def compare(self):
        metrics, inf_nan_consistency = self._compute_ratio()
        compare_result = self._post_compare(metrics, inf_nan_consistency)
        return compare_result
    
    def _get_and_convert_values(self, column_name):
        npu_value = self._get_and_convert_value(self.row_npu, column_name, "NPU")
        gpu_value = self._get_and_convert_value(self.row_gpu, column_name, "GPU")
        return npu_value, gpu_value

    def _post_compare(self, metrics, inf_nan_consistency):
        metrics = self._get_status(metrics, inf_nan_consistency)
        metrics.update({'compare_algorithm': self.compare_algorithm})
        self.compare_column.update(metrics)
        compare_result = metrics.get('compare_result')
        return compare_result


# coding=utf-8
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

"""
Function:
FusionOpComparison class. This class mainly involves the compare function.
"""
import collections

from msprobe.msaccucmp.vector_cmp.fusion_manager import fusion_rule_parser
from msprobe.msaccucmp.cmp_utils import utils, utils_type
from msprobe.msaccucmp.cmp_utils import log
from msprobe.msaccucmp.vector_cmp.fusion_manager.compare_npu_vs_npu import NpuVsNpuComparison
from msprobe.msaccucmp.vector_cmp.fusion_manager import compare_result
from msprobe.msaccucmp.dump_parse.dump import DumpType
from msprobe.msaccucmp.overflow.overflow_detection import OverflowDetection
from msprobe.msaccucmp.vector_cmp.range_manager.range_manager import RangeManager
from msprobe.msaccucmp.vector_cmp.fusion_manager.fusion_op import FusionOp
from msprobe.msaccucmp.vector_cmp.fusion_manager.fusion_op import Tensor
from msprobe.msaccucmp.cmp_utils.constant.const_manager import ConstManager
from msprobe.msaccucmp.conversion.tensor_conversion import TensorConversion
from msprobe.msaccucmp.cmp_utils.constant.compare_error import CompareError
from msprobe.msaccucmp.dump_parse.dump_data_object import DumpDataObj
from msprobe.msaccucmp.dump_parse import dump_utils
from msprobe.msaccucmp.cmp_utils import common


class FusionOpComparison:
    """
    The class for fusion op compare
    """
    TIMESTAMP_INDEX = 0
    ORIGINAL_NAMES_INDEX = 2

    def __init__(self: any, *args: any) -> None:
        fusion_op_name, compare_rule, compare_data, format_manager, arguments = args
        self.compare_data = compare_data
        self.compare_rule = compare_rule
        self.format_manager = format_manager
        self.algorithm_manager = arguments.get('algorithm_manager')
        self.fusion_op_list = self.compare_rule.fusion_info.fusion_op_name_to_op_map[fusion_op_name]
        self.left_dump_file_path = ''
        self.left_dump_data = None
        self.overflow_detection = arguments.get('overflow_detection', False)
        self.is_ground_truth_gpu_or_cpu = \
            True if utils.dump_path_contains_npy(arguments.get("golden_dump_path")) else False
        self.is_my_dump_gpu_or_cpu = \
            True if utils.dump_path_contains_npy(arguments.get("my_dump_path")) else False
        self.op_name_origin_output_index_map = {}
        self.max_cmp_size = arguments.get('max_cmp_size')

    @staticmethod
    def _check_dequant_op(fusion_op: FusionOp) -> bool:
        if fusion_op.op_type == 'AscendDequant':
            return True
        for suffix in ConstManager.DEQUANT_OP_NANE_SUFFIX_LIST:
            if fusion_op.op_name.endswith(suffix):
                return True
        return False

    @staticmethod
    def _handle_error_msg(compare_error: CompareError, fusion_op: FusionOp, tensor_id: str) -> str:
        if compare_error.code == CompareError.MSACCUCMP_NO_DUMP_FILE_ERROR:
            error_msg = [log.print_no_right_dump_file_error(fusion_op.op_name, tensor_id)]
        elif compare_error.code == CompareError.MSACCUCMP_INVALID_FRACTAL_NZ_DUMP_DATA_ERROR:
            error_msg = [log.print_invalid_nz_dump_data(compare_error.message, op_name=fusion_op.op_name)]
        else:
            error_msg = [compare_error.message]
        return error_msg

    def sort_l1_fusion_dump_file(self: any) -> list:
        """
        Sort l1 fusion dump file by timestamp
        :return: the sorted dump file by timestamp
        """
        timestamp_list = []
        for fusion_op in self.fusion_op_list:
            try:
                file_path_list, data_mode = self.compare_data.left_dump_info.get_op_dump_file(
                    fusion_op.op_name, print_log=False)
            except CompareError:
                continue
            finally:
                pass
            file_path = file_path_list[-1]
            if data_mode == ConstManager.NORMAL_MODE:
                timestamp = dump_utils.get_normal_timestamp(file_path)
            else:
                timestamp = dump_utils.get_ffts_timestamp(file_path)
            original_names = utils.get_string_from_list(
                fusion_op.attr.original_op_names)
            timestamp_list.append([timestamp, fusion_op, original_names])
        # sort by timestamp
        sort_timestamp_list = sorted(timestamp_list, key=lambda s: s[self.TIMESTAMP_INDEX])
        return sort_timestamp_list

    def get_right_dump_data(self: any, fusion_op: FusionOp, index: int, is_input: bool = False,
                            parse: bool = True, tensor_id: str = None) -> Tensor:
        """
        Get the right dump file path and data by left dump data
        :param fusion_op: the fusion op
        :param index: the index
        :param is_input: the left data is input or not
        :param parse: need parse dump data or not
        :param tensor_id: tensor index
        :return tensor and dump data
        """
        compare_fusion_op = fusion_op
        compare_index = index
        if is_input:
            compare_fusion_op, compare_index = self._find_pre_op(fusion_op, index)
        while compare_fusion_op.op_type in ConstManager.SPECIAL_OPS_TYPE:
            # Assume that the input and output indexes of the special op are in one-to-one correspondence.
            compare_fusion_op, compare_index = self._find_pre_op(compare_fusion_op, compare_index)

        origin_tensor = compare_fusion_op.get_origin_tensor(compare_index)
        if is_input and tensor_id:
            self.op_name_origin_output_index_map[tensor_id] = (compare_fusion_op.op_name, compare_index)
        left_data_type = ConstManager.INPUT if is_input else ConstManager.OUTPUT
        log.print_info_log('[%s] Left(%s:%s:%d) <======> Right(%s:%s:%d)'
                           % (fusion_op.op_name, fusion_op.op_name, left_data_type, index, origin_tensor.name,
                              ConstManager.OUTPUT, origin_tensor.index))
        if not parse:
            index = None if self.compare_data.right_dump_info.type == DumpType.Offline else origin_tensor.index
            dump_file_list, _ = self.compare_data.right_dump_info.get_op_dump_file(origin_tensor.name, index)
            dump_file_path = dump_file_list[-1]
            origin_tensor.set_path(dump_file_path)
            return origin_tensor
        dump_file_path, dump_data = self.compare_data.get_right_dump_data(
            origin_tensor.name, origin_tensor.index)
        origin_tensor.set_path(dump_file_path)
        if self.compare_data.right_dump_info.type == DumpType.Offline:
            if origin_tensor.index >= len(dump_data.output_data):
                log.print_out_of_range_error(fusion_op.op_name, "output", origin_tensor.index,
                                             '[0, %d)' % len(dump_data.output_data))
                raise CompareError(CompareError.MSACCUCMP_INDEX_OUT_OF_BOUNDS_ERROR)

            origin_tensor.set_data(dump_data.output_data[origin_tensor.index])
            origin_tensor.tensor_format = \
                common.get_format_string(dump_data.output_data[origin_tensor.index].tensor_format)
            origin_tensor.shape = dump_data.output_data[origin_tensor.index].shape

        else:
            origin_tensor.set_data(dump_data.output_data[0])
        return origin_tensor

    def compare(self: any) -> (int, bool, list):
        """
        Compare for one the fusion op
        :return error_code:VectorComparisonErrorCode
        :return dump_match: True, at least one operator match;False, no operator match
        :return result: the compare result by the fusion op
        """
        ret = CompareError.MSACCUCMP_NONE_ERROR
        dump_match = False
        result = None

        try:
            if self.compare_rule.quant_fusion_rule_file_path == '' \
                    and self.compare_rule.fusion_json_file_path == '':
                npu_vs_npu_comparison = NpuVsNpuComparison(self.compare_data, self.fusion_op_list,
                                                           self.algorithm_manager, self.overflow_detection)
                return npu_vs_npu_comparison.compare()
        except (OSError, SystemError, ValueError, TypeError, RuntimeError, MemoryError,
                AttributeError) as error:
            log.print_error_log('Failed to compare %s. %s' % (self.fusion_op_list[0].op_name, error))
            return CompareError.MSACCUCMP_UNKNOWN_ERROR, False, []

        # 1. get the fusion op relation
        relation = fusion_rule_parser.get_relation_for_fusion(self.fusion_op_list)
        # 2. get the map for {original_op_names, op_list}
        right_to_left_map = fusion_rule_parser.make_right_to_left_multi_map(self.fusion_op_list)
        # 3. compare by relation
        if relation in (utils_type.FusionRelation.OneToOne, utils_type.FusionRelation.MultiToOne):
            dump_match, result, ret = self._compare_for_any_to_one()
        elif relation in (utils_type.FusionRelation.OneToMulti, utils_type.FusionRelation.MultiToMulti):
            dump_match, result, ret = self._compare_for_any_to_multi(right_to_left_map)
        elif relation == utils_type.FusionRelation.L1Fusion:
            dump_match, result, ret = self._compare_for_l1_fusion(right_to_left_map)
        return ret, dump_match, result

    def make_gpu_and_npu_mapping_table(self: any) -> list:
        """
        Generate table content by op name
        :return table content
        """
        table_content = []
        for fusion_op in self.fusion_op_list:
            ok, my_output_dump_path = self._get_my_output_dump_path(fusion_op, table_content)
            if not ok:
                continue
            ok, my_output_dump_data = self._parse_dump_file(fusion_op, my_output_dump_path, table_content)
            if not ok:
                continue
            self._make_mapping_by_input(fusion_op, my_output_dump_data.input_data, my_output_dump_path, table_content)
            self._make_mapping_by_output(fusion_op, my_output_dump_data.output_data, my_output_dump_path, table_content)
        return table_content

    def _compare_by_operator(self: any, fusion_op: FusionOp) -> list:
        """
        Compare for one operator
        """
        log.print_start_to_compare_op(fusion_op.op_name)
        try:
            self.left_dump_file_path, self.left_dump_data = \
                self.compare_data.get_left_dump_data(fusion_op.op_name)
        except CompareError as error:
            error.message = log.print_no_left_dump_file_error(fusion_op.op_name, fusion_op.op_type)
            raise error
        finally:
            pass
        compare_vector_result = []
        if not self._check_dequant_op(fusion_op):
            _, result = self._compare_by_tensor(fusion_op, True)
            compare_vector_result += result
        if fusion_op.op_type != 'AscendQuant':
            _, result = self._compare_by_tensor(fusion_op, False)
            compare_vector_result += result
        return compare_vector_result

    def _compare_for_any_to_one(self: any) -> (bool, list, int):
        """
        Compare for any to one relation fusion op
        """
        dump_match = False
        cmp_result_list = []
        # compare each fusion op
        fusion_op_result = compare_result.FusionOpComResult(self.algorithm_manager,
                                                            overflow_detection=self.overflow_detection,
                                                            dump_is_cpu_or_gpu_data=[self.is_ground_truth_gpu_or_cpu,
                                                                                     self.is_my_dump_gpu_or_cpu])
        ret = CompareError.MSACCUCMP_NONE_ERROR

        for fusion_op in self.fusion_op_list:
            single_op_cmp_result = compare_result.SingleOpCmpResult()
            # skip not support compare op
            if fusion_op.attr.quant_filter:
                log.print_skip_quant_info(fusion_op.op_name)
                continue
            result = None
            # 1. compare op
            error_msg = []
            try:
                result = self._compare_by_operator(fusion_op)
            except CompareError as compare_error:
                ret = compare_error.code
                error_msg.append(compare_error.message)
            if result:
                dump_match = True
            # 2. make compare result to string
            _result = fusion_op_result.get_result(fusion_op, result, error_msg)

            result_info = utils.ResultInfo(
                fusion_op.op_name, dump_match, _result.result_list, ret,
                fusion_op.input_list, _result.input_result_list, _result.output_result_list,
                _result.is_ffts, self.op_name_origin_output_index_map, False)

            single_op_cmp_result.update_attr(result_info)
            cmp_result_list.append(single_op_cmp_result)
        return dump_match, cmp_result_list, ret

    def _compare_for_any_to_multi(self: any, right_to_left_map: dict) -> (bool, list, int):
        """
        Compare for any to multi relation fusion op
        """
        dump_match = False
        has_result_for_any_to_multi = False
        error_msg = []
        cmp_result_list = []
        ret = CompareError.MSACCUCMP_NONE_ERROR
        fusion_op_result = compare_result.FusionOpComResult(self.algorithm_manager, right_to_left_map,
                                                            self.overflow_detection,
                                                            dump_is_cpu_or_gpu_data=[self.is_ground_truth_gpu_or_cpu,
                                                                                     self.is_my_dump_gpu_or_cpu])
        # compare each fusion op
        for fusion_op in self.fusion_op_list:
            single_op_cmp_result = compare_result.SingleOpCmpResult()
            if fusion_op.attr.quant_filter:
                log.print_skip_quant_info(fusion_op.op_name)
                continue
            result = None
            if fusion_op.is_inner_node():
                warn_message = '[%s] The op is inner node for multi to multi relation.Skip the op.' \
                               % fusion_op.op_name
                log.print_warn_log(warn_message)
                error_msg.append(warn_message)
                continue
            # 1. compare op
            try:
                result = self._compare_by_operator(fusion_op)
            except CompareError as compare_error:
                ret = compare_error.code
                error_msg.append(compare_error.message)
                if ret == CompareError.MSACCUCMP_NO_DUMP_FILE_ERROR:
                    continue
            finally:
                pass
            has_result_for_any_to_multi = True
            dump_match = True
            error_msg.clear()
            # 2. write compare result to file
            _result = fusion_op_result.get_result(fusion_op, result, error_msg)

            result_info = utils.ResultInfo(
                fusion_op.op_name, dump_match, _result.result_list, ret,
                fusion_op.input_list, _result.input_result_list, _result.output_result_list,
                _result.is_ffts, self.op_name_origin_output_index_map, False)

            single_op_cmp_result.update_attr(result_info)
            cmp_result_list.append(single_op_cmp_result)
        # no result for any to multi, write empty result to file
        if not has_result_for_any_to_multi:
            single_op_cmp_result = compare_result.SingleOpCmpResult()
            _result = fusion_op_result.get_result(self.fusion_op_list[0], None, error_msg)
            result_info = utils.ResultInfo(
                self.fusion_op_list[0].op_name, dump_match, _result.result_list, ret,
                self.fusion_op_list[0].input_list, _result.input_result_list, _result.output_result_list,
                _result.is_ffts, self.op_name_origin_output_index_map, False)
            single_op_cmp_result.update_attr(result_info)
            cmp_result_list.append(single_op_cmp_result)
        return dump_match, cmp_result_list, ret

    def _compare_for_l1_fusion_not_timestamp(self, fusion_op_result):
        error_msg = []

        single_op_cmp_result = compare_result.SingleOpCmpResult()
        error_msg.append("The fusion operator list is empty")
        _result = fusion_op_result.get_result(self.fusion_op_list[0], None, error_msg)
        result_info = utils.ResultInfo(self.fusion_op_list[0].op_name, False, _result.result_list,
                                       CompareError.MSACCUCMP_NO_DUMP_FILE_ERROR, self.fusion_op_list[0].input_list,
                                       _result.input_result_list, _result.output_result_list, _result.is_ffts,
                                       self.op_name_origin_output_index_map, False)
        single_op_cmp_result.update_attr(result_info)
        return False, [single_op_cmp_result], CompareError.MSACCUCMP_NO_DUMP_FILE_ERROR

    def _compare_for_l1_fusion(self: any, right_to_left_map: dict) -> (bool, list, int):
        timestamp_list = self.sort_l1_fusion_dump_file()
        error_msg = []
        fusion_op_result = compare_result.FusionOpComResult(self.algorithm_manager, right_to_left_map,
                                                            self.overflow_detection,
                                                            dump_is_cpu_or_gpu_data=[self.is_ground_truth_gpu_or_cpu,
                                                                                     self.is_my_dump_gpu_or_cpu])

        if not timestamp_list:
            return self._compare_for_l1_fusion_not_timestamp(fusion_op_result)

        # if the dump data is only input,
        # the dump file of min and max timestamp for input is the same
        # if the dump data is only output,
        # the dump file of max timestamp for output is complete
        # if the dump data contains input and output,
        # the dump file of max timestamp for output is complete
        # the dump file of min timestamp for input is complete
        l1_fusion_list = []
        if timestamp_list[-1][self.ORIGINAL_NAMES_INDEX] != timestamp_list[0][self.ORIGINAL_NAMES_INDEX]:
            l1_fusion_list.append(timestamp_list[0][ConstManager.FUSION_OP_INDEX])
        l1_fusion_list.append(timestamp_list[-1][ConstManager.FUSION_OP_INDEX])
        dump_match = False
        has_result_for_any_to_multi = False
        ret = CompareError.MSACCUCMP_NONE_ERROR
        cmp_result_list = []

        # compare each fusion op
        for fusion_op in l1_fusion_list:
            single_op_cmp_result = compare_result.SingleOpCmpResult()
            if fusion_op.attr.quant_filter:
                log.print_skip_quant_info(fusion_op.op_name)
                continue

            result = None
            # 1. compare op
            try:
                result = self._compare_by_operator(fusion_op)
            except CompareError as compare_error:
                ret = compare_error.code
                error_msg.append(compare_error.message)
                if ret == CompareError.MSACCUCMP_NO_DUMP_FILE_ERROR:
                    continue
            error_msg.clear()
            has_result_for_any_to_multi = True
            dump_match = True
            # 2. write compare result to file
            _result = fusion_op_result.get_result(fusion_op, result, error_msg)

            result_info = utils.ResultInfo(fusion_op.op_name, dump_match, _result.result_list, ret,
                fusion_op.input_list, _result.input_result_list, _result.output_result_list,
                _result.is_ffts, self.op_name_origin_output_index_map, False)

            single_op_cmp_result.update_attr(result_info)
            cmp_result_list.append(single_op_cmp_result)

        # no result for any to multi, write empty result to file
        if not has_result_for_any_to_multi:
            single_op_cmp_result = compare_result.SingleOpCmpResult()
            _result = fusion_op_result.get_result(self.fusion_op_list[0], None, error_msg)

            result_info = utils.ResultInfo(
                self.fusion_op_list[0].op_name, dump_match, _result.result_list, ret,
                self.fusion_op_list[0].input_list, _result.input_result_list, _result.output_result_list,
                _result.is_ffts, self.op_name_origin_output_index_map, False)

            single_op_cmp_result.update_attr(result_info)
            cmp_result_list.append(single_op_cmp_result)

        return dump_match, cmp_result_list, ret

    def _find_pre_op(self: any, fusion_op: FusionOp, index: int = 0) -> any:
        input_tensor = fusion_op.get_input_tensor(index)
        _, compare_fusion_op = \
            self.compare_rule.fusion_info.get_fusion_op_list(input_tensor.name)
        compare_index = input_tensor.index
        return compare_fusion_op, compare_index

    def _compare_by_one_tensor(self: any, fusion_op: FusionOp, index: int,
                               is_input: bool, tensor: any, tensor_id: str) -> (list, list, list):
        # 1. get ground truth dump data by original op name and index
        ground_truth_tensor = self.get_right_dump_data(fusion_op, index, is_input, tensor_id=tensor_id)
        if tensor.original_shape:
            ground_truth_tensor.shape = tensor.original_shape
        # 2. deserialize output data to array
        tensor_conversion = TensorConversion(fusion_op, self.format_manager, is_detail=False)
        my_output_array, ground_truth_array, my_output_shape = \
            tensor_conversion.get_my_output_and_ground_truth_data(self.compare_data, tensor, ground_truth_tensor)
        my_output_dtype = my_output_array.dtype
        ground_truth_dtype = ground_truth_array.dtype

        # 3. compare by support algorithm
        result, compare_fail_message = self.algorithm_manager.compare(
            my_output_array, ground_truth_array,
            {'my_output_dump_file': self.left_dump_file_path,
             'ground_truth_dump_file': ground_truth_tensor.path,
             'shape_type': utils.get_shape_type(my_output_shape)},
            self.max_cmp_size)

        return result, compare_fail_message, [my_output_shape, my_output_dtype, ground_truth_dtype]

    def _compare_by_tensor(self: any, fusion_op: FusionOp, is_input: bool) -> (bool, list):
        tensor_list = self.left_dump_data.input_data if is_input else self.left_dump_data.output_data
        match = False
        tensor_result_list = []
        # compare each tensor
        for (index, tensor) in enumerate(tensor_list):
            result = self._compare(fusion_op, index, is_input, tensor)
            # Check whether the current input/output data overflows
            overflow_result = ''
            if self.overflow_detection:
                overflow_result = OverflowDetection.process_model_overflow_detection(fusion_op.op_name, index, is_input,
                                                                                     tensor)
            match = match or result.match
            # 4. merge result
            tensor_info = {
                "tensor_id": result.tensor_id,
                "shape": result.shape,
                "op_type": fusion_op.op_type,
                "my_output_dtype": result.my_output_dtype,
                "ground_truth_dtype": result.ground_truth_dtype,
                "my_output_address": result.my_output_address,
                "ground_truth_address": result.ground_truth_address
            }
            tensor_result = compare_result.TensorResult(tensor_info, [result.algorithm_result, overflow_result],
                                                        result.error_msg, tensor.is_ffts)
            tensor_result_list.append(tensor_result)
        return match, tensor_result_list

    def _compare(self: any, fusion_op: FusionOp, index: int, is_input: bool, tensor: any) -> tuple:
        tensor_type = ConstManager.INPUT if is_input else ConstManager.OUTPUT
        tensor_id = "%s:%s:%d" % (fusion_op.op_name, tensor_type, index)
        match = True
        my_output_dtype = "NaN"
        ground_truth_dtype = "NaN"
        try:
            algorithm_result, error_msg, [shape, my_output_dtype, ground_truth_dtype] = self._compare_by_one_tensor(
                fusion_op, index, is_input, tensor, tensor_id)
        except CompareError as compare_error:
            error_msg = self._handle_error_msg(compare_error, fusion_op, tensor_id)
            algorithm_result = self.algorithm_manager.make_nan_result()
            shape = tensor.shape
            match = False

        CompareResult = collections.namedtuple(
            'Result',
            ['tensor_id', 'shape', 'algorithm_result', 'error_msg', 'match', 'my_output_dtype', 'ground_truth_dtype',
             'my_output_address', 'ground_truth_address'])
        result = CompareResult(tensor_id, shape, algorithm_result, error_msg, match, my_output_dtype,
                               ground_truth_dtype,
                               my_output_address=utils.get_address_from_tensor(tensor),
                               ground_truth_address=ConstManager.NAN)
        return result

    def _get_my_output_dump_path(self: any, fusion_op: FusionOp, table_content: list) -> (bool, str):
        output_index = None
        left_dump_info = self.compare_data.left_dump_info
        if left_dump_info.type == DumpType.Quant:
            output_index = 0
        try:
            return True, self.compare_data.left_dump_info.get_op_dump_file(fusion_op.op_name, output_index)[0][-1]
        except IndexError as e:
            log.print_error_log("index out of bounds error when get op dump file, please check.")
            raise CompareError(CompareError.MSACCUCMP_INDEX_OUT_OF_BOUNDS_ERROR) from e
        except (OSError, SystemError, ValueError, TypeError, RuntimeError, MemoryError,
                AttributeError, CompareError):
            message = '[{0}] There is no left dump file for the op "{0}".'.format(fusion_op.op_name)
            log.print_warn_log(message)
            table_content.append(self._make_mapping_table_content(
                ConstManager.NAN, ConstManager.NAN, ConstManager.NAN, fusion_op))
        return False, ''

    def _get_ground_truth_dump_path(self: any, fusion_op: FusionOp, index: int, is_input: bool) -> (bool, str):
        ground_truth_tensor = None
        try:
            ground_truth_tensor = self.get_right_dump_data(fusion_op, index, is_input, parse=False)
        except (OSError, SystemError, ValueError, TypeError, RuntimeError, MemoryError,
                AttributeError, CompareError):
            message = '[{0}] There is no right dump file for the op "{0}".'.format(fusion_op.op_name)
            log.print_warn_log(message)

        if ground_truth_tensor:
            return ground_truth_tensor.path
        else:
            return ConstManager.NAN

    def _parse_dump_file(self: any, fusion_op: FusionOp, my_output_dump_path: str,
                         table_content: list) -> (bool, DumpDataObj):
        try:
            return True, dump_utils.parse_dump_file(my_output_dump_path, self.compare_data.dump_version)
        except (OSError, SystemError, ValueError, TypeError, RuntimeError, MemoryError,
                AttributeError, CompareError):
            log.print_error_log("{} file parse error".format(my_output_dump_path))
            table_content.append(self._make_mapping_table_content(
                ConstManager.NAN, my_output_dump_path, ConstManager.NAN, fusion_op))
        finally:
            pass
        return False, DumpDataObj()

    def _make_mapping_table_content(self: any, tensor_id: str, my_output_dump_path: str, ground_truth_dump_path: str,
                                    fusion_op: FusionOp) -> list:
        relation = fusion_rule_parser.get_relation_for_fusion(self.fusion_op_list)
        ground_truth_to_my_output_map = fusion_rule_parser.make_right_to_left_multi_map(self.fusion_op_list)
        if relation in (utils_type.FusionRelation.OneToOne, utils_type.FusionRelation.MultiToOne):
            ground_truth_op_str = utils.get_string_from_list(fusion_op.attr.original_op_names)
            if ground_truth_op_str == "":
                ground_truth_op_str = "*"
            my_output_op_str = fusion_op.op_name
        else:
            my_output_op_str, ground_truth_op_str = fusion_rule_parser.make_left_and_right_string(
                ground_truth_to_my_output_map)
        if fusion_op.op_type in [ConstManager.LEFT_TYPE, ConstManager.RIGHT_TYPE]:
            op_type = dump_utils.get_op_type_from_file_name(my_output_dump_path)
        else:
            op_type = fusion_op.op_type
        row_content = [
            str(fusion_op.op_id), op_type, my_output_op_str, ground_truth_op_str,
            tensor_id, my_output_dump_path, ground_truth_dump_path
        ]
        RangeManager.adjust_data(row_content, fusion_op.attr.get_op_sequence())
        return row_content

    def _make_mapping_by_input(self: any, fusion_op: FusionOp, input_tensor: any,
                               my_output_dump_path: str, table_content: list) -> None:
        if not self._check_dequant_op(fusion_op):
            for index, _ in enumerate(input_tensor):
                tensor_id = "%s:%s:%d" % (fusion_op.op_name, ConstManager.INPUT, index)
                ground_truth_dump_path = self._get_ground_truth_dump_path(fusion_op, index, is_input=True)
                table_content.append(self._make_mapping_table_content(
                    tensor_id, my_output_dump_path, ground_truth_dump_path, fusion_op))

    def _make_mapping_by_output(self: any, fusion_op: FusionOp, output_tensor: any,
                                my_output_dump_path: str, table_content: list) -> None:
        if fusion_op.op_type != 'AscendQuant':
            for index, _ in enumerate(output_tensor):
                tensor_id = "%s:%s:%d" % (fusion_op.op_name, ConstManager.OUTPUT, index)
                ground_truth_dump_path = self._get_ground_truth_dump_path(fusion_op, index, is_input=False)
                table_content.append(self._make_mapping_table_content(
                    tensor_id, my_output_dump_path, ground_truth_dump_path, fusion_op))



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, Dict, List

from msprobe.core.common.const import Const, MonitorConst
from msprobe.core.common.framework_adapter import FmkAdp
from msprobe.core.common.log import logger
from msprobe.core.monitor_v2.base import BaseMonitorV2
from msprobe.core.monitor_v2.utils import build_param2name, get_vpp_stage_from_tag


class ParamMonitorV2(BaseMonitorV2, ABC):
    """
    Parameter distribution monitor that captures params before/after optimizer step.
    """

    def __init__(self) -> None:
        super().__init__()
        self._optimizer: Any = None
        self._model: Any = None
        self._param2name: Dict[Any, str] = {}
        self._eps: float = 1e-8
        self.param_distribution: bool = True
        self._step_hook_handles: List[Any] = []
        self._orig_step = None
        self._step_patched = False

    @abstractmethod
    def _compute_metrics(self, tag2tensor: Dict[str, Any]) -> Dict[str, Dict[str, Any]]:
        ...

    def set_config(self, config: Dict[str, Any]) -> None:
        super().set_config(config)
        self.config = config
        self._eps = float(config.get("eps", 1e-8))
        self.param_distribution = bool(config.get("param_distribution", True))

    def start(self, model: Any = None, optimizer: Any = None, **context: Any) -> None:
        self.set_context(**context)
        if self._optimizer is not None:
            return
        optimizer = optimizer or self._context.get("optimizer")
        model = model or self._context.get("model")
        if optimizer is None or model is None:
            raise ValueError("Model and optimizer must be provided to start ParamMonitorV2")
        self._optimizer = optimizer
        self._model = model
        self._param2name = build_param2name(model=model, optimizer=optimizer)
        self._patch_optimizer()

    def stop(self) -> None:
        self._restore_optimizer()
        self._optimizer = None
        self._model = None
        self._param2name.clear()
        self._rows = []

    def _patch_optimizer(self) -> None:
        if not self.param_distribution or self._optimizer is None:
            return
        pre_hook = getattr(self._optimizer, "register_step_pre_hook", None)
        post_hook = getattr(self._optimizer, "register_step_post_hook", None)
        if callable(pre_hook) and callable(post_hook):
            self._step_hook_handles.append(pre_hook(self._on_pre_step))
            self._step_hook_handles.append(post_hook(self._on_post_step))
            return
        if self._step_patched or not hasattr(self._optimizer, "step"):
            return
        self._orig_step = self._optimizer.step

        def patched_step(*args: Any, **kwargs: Any):
            self._record_param_distribution(MonitorConst.PRE_PARAM)
            out = self._orig_step(*args, **kwargs)
            self._record_param_distribution(MonitorConst.POST_PARAM)
            return out

        self._optimizer.step = patched_step  # type: ignore[assignment]
        self._step_patched = True

    def _restore_optimizer(self) -> None:
        for handle in self._step_hook_handles:
            try:
                handle.remove()
            except Exception:
                pass
        self._step_hook_handles.clear()
        if self._step_patched and self._optimizer is not None and self._orig_step is not None:
            self._optimizer.step = self._orig_step  # type: ignore[assignment]
        self._orig_step = None
        self._step_patched = False

    def _on_pre_step(self, optimizer: Any, args: Any, kwargs: Any) -> None:
        self._record_param_distribution(MonitorConst.PRE_PARAM)

    def _on_post_step(self, optimizer: Any, args: Any, kwargs: Any) -> None:
        self._record_param_distribution(MonitorConst.POST_PARAM)

    def _record_param_distribution(self, scope: str) -> None:
        if not self.param_distribution or not self._param2name:
            return
        tag2tensor: Dict[str, Any] = {}
        for param, name in self._param2name.items():
            try:
                if param.numel() == 0:
                    continue
            except Exception:
                continue
            if not FmkAdp.is_tensor(param):
                continue
            tensor = param
            if tensor.__class__.__name__ == "DTensor":
                try:
                    tensor = tensor.to_local()
                except Exception as exc:
                    logger.warning(f"[monitor_v2] Failed to convert DTensor param to local: {exc}")
                    continue
            tag2tensor[name] = tensor
        if not tag2tensor:
            return
        stats = self._compute_metrics(tag2tensor)
        for tag, op_dict in stats.items():
            vpp_stage = get_vpp_stage_from_tag(tag)
            row = {
                "module_name": tag,
                "scope": scope,
                "stats": {op: self._detach_stat(val) for op, val in op_dict.items()},
            }
            if vpp_stage is not None:
                row["vpp_stage"] = vpp_stage
            else:
                row["vpp_stage"] = MonitorConst.DEFAULT_STAGE
            self._rows.append(row)

    def _detach_stat(self, value: Any) -> Any:
        if value is None or not FmkAdp.is_tensor(value):
            return value
        if FmkAdp.fmk == Const.PT_FRAMEWORK:
            return value.detach()
        return value


class PyTorchParamMonitorV2(ParamMonitorV2):
    def _compute_metrics(self, tag2tensor: Dict[str, Any]) -> Dict[str, Dict[str, Any]]:
        from msprobe.pytorch.monitor.module_metric import get_metrics

        return get_metrics(self._ops, tag2tensor, self._eps, {})


class MindSporeParamMonitorV2(ParamMonitorV2):
    def _compute_metrics(self, tag2tensor: Dict[str, Any]) -> Dict[str, Dict[str, Any]]:
        from msprobe.mindspore.monitor.utils import get_metrics

        return get_metrics(self._ops, tag2tensor, self._eps, {})



"""
monitor_v2: next-generation monitoring core for unifying
PyTorch / MindSpore monitor logic behind a single, clean abstraction.

Goal:
  - keep production behavior stable while iterating;
  - migrate mature logic from existing monitors in small steps;
  - enforce clearer layering and performance constraints.

Current structure:
  base.py
    - BaseMonitorV2: minimal shared monitor scaffold.
  trainer.py
    - TrainerMonitorV2: orchestrator for step gating and writer integration.
  factory.py
    - MonitorFactory: registry of framework-specific monitor classes.
  writer.py
    - CSVWriterV2: CSV output for per-step monitor data.
"""
from .trainer import TrainerMonitorV2



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import functools
import inspect
import os

import torch
import torch.distributed as dist

from msprobe.core.common.const import Const
from msprobe.core.common.file_utils import load_yaml
from msprobe.core.common.runtime import Runtime
from msprobe.core.dump.api_dump.api_registry import ApiRegistry
from msprobe.pytorch.common.log import logger
from msprobe.pytorch.common.utils import (
    torch_without_guard_version,
    is_gpu,
    torch_device_guard,
    parameter_adapter
)
from msprobe.pytorch.dump.function_factory import npu_custom_functions
from msprobe.pytorch.dump.api_dump.hook_module import HOOKModule
from msprobe.pytorch.dump.api_dump.utils import dynamic_import_op

try:
    import mindspeed.ops
except ImportError:
    mindspeed_enable = False
else:
    mindspeed_enable = True

torch_version_above_2 = torch.__version__.split('+')[0] > '2.0'

_supported_api_list_path = (os.path.join(os.path.dirname(os.path.realpath(__file__)), Const.SUPPORT_API_FILE_NAME),)
_cuda_func_mapping = {"npu_fusion_attention": "gpu_fusion_attention"}
dist_data_collect_func = {}
dist_batch_data_collect_func = []

_api_types = {
    Const.PT_FRAMEWORK: {
        Const.PT_API_TYPE_FUNCTIONAL: ((torch.nn.functional,), (torch.nn.functional,)),
        Const.PT_API_TYPE_TENSOR: ((torch.Tensor,), (torch.Tensor,)),
        Const.PT_API_TYPE_TORCH: ((torch,), (torch,)),
        Const.PT_API_TYPE_VF: ((torch._C._VariableFunctionsClass,), (torch._VF,)),
        Const.PT_API_TYPE_DIST: ((dist,), (dist,)),
        Const.PT_API_TYPE_DIST_C10D: ((dist.distributed_c10d,), (dist.distributed_c10d,))
    }
}
if not is_gpu:
    import torch_npu

    if torch_without_guard_version:
        _api_types.get(Const.PT_FRAMEWORK).update(
            {
                Const.PT_API_TYPE_NPU: ((torch.ops.npu, torch_npu), (torch_npu, torch.ops.npu)),
            }
        )
    else:
        _api_types.get(Const.PT_FRAMEWORK).update(
            {Const.PT_API_TYPE_NPU: ((torch_npu._C._VariableFunctionsClass,), (torch_npu,))}
        )
        _api_types.get(Const.PT_FRAMEWORK).update(
            {
                Const.PT_API_TYPE_NPU_DIST: (
                    (torch_npu.distributed,),
                    (torch_npu.distributed, torch_npu.distributed.distributed_c10d)
                )
            }
        )
    if mindspeed_enable:
        _api_types.get(Const.PT_FRAMEWORK).update({Const.PT_API_TYPE_MINDSPEED: ((mindspeed.ops,), (mindspeed.ops,))})
        mindspeed_op_list = load_yaml(_supported_api_list_path[0]).get(Const.PT_API_TYPE_MINDSPEED)
        mindspeed_op_file_list = [op.split(Const.SEP)[0] + Const.PY_SUFFIX for op in mindspeed_op_list]
        dynamic_import_op(mindspeed.ops, mindspeed_op_file_list)

_inner_used_api = {
    Const.PT_FRAMEWORK + Const.SEP + Const.PT_API_TYPE_TENSOR: (
        torch.Tensor, "view_as"
    )
}


def reset_dist_collect_func():
    global dist_data_collect_func, dist_batch_data_collect_func
    dist_data_collect_func.clear()
    dist_batch_data_collect_func.clear()


@parameter_adapter
def tensor_module_forward(module, *args, **kwargs):
    return module.api_func(*args, **kwargs)


def dist_module_forward(module, *args, **kwargs):
    handle = module.api_func(*args, **kwargs)
    try:
        bound = inspect.signature(module.api_func).bind(*args, **kwargs)
        bound.apply_defaults()
        use_async_op_flag = bound.arguments.get("async_op", False)
    except Exception as e:
        use_async_op_flag = False
        logger.warning(f"fail to get dist api's func signature because {e}, no wait")

    def create_async_callback_func(catch_func):
        full_name = module.full_forward_name if hasattr(module, "full_forward_name") else None

        def store_data():
            catch_func(module, full_name, args, kwargs, handle)

        return store_data

    if Runtime.is_running and (use_async_op_flag or module.api_name in ['isend', 'irecv']):
        dist_data_collect_func[handle] = create_async_callback_func(module.distributed_forward_hook)
    if Runtime.is_running and module.api_name == 'batch_isend_irecv':
        dist_batch_data_collect_func.append([handle, create_async_callback_func(module.distributed_forward_hook)])
    return handle


def redirect_wait():
    if hasattr(dist, "Work"):
        from torch.distributed import Work
    else:
        from torch._C._distributed_c10d import Work
    origin_wait = Work.wait

    def wrapped_wait(work):
        def wrapped_wait(*args, **kwargs):
            origin_wait(*args, **kwargs)
            if args[0] in dist_data_collect_func:
                store_func = dist_data_collect_func.pop(args[0])
                store_func()
                return
            remove_value = None
            for value in dist_batch_data_collect_func:
                if args[0] in value[0]:
                    value[0].remove(args[0])
                    if len(value[0]) == 0:
                        store_func = value[1]
                        store_func()
                        remove_value = value
                    break
            if remove_value:
                dist_batch_data_collect_func.remove(remove_value)

        return wrapped_wait

    Work.wait = wrapped_wait(Work)


def npu_module_forward(module, *args, **kwargs):
    if not module.need_hook:
        if module.api_name not in npu_custom_functions:
            raise Exception(f'There is not bench function {module.api_name}')
        if module.device == Const.CUDA_LOWERCASE:
            module.api_name = _cuda_func_mapping.get(module.api_name, module.api_name)
        if module.device in [Const.CUDA_LOWERCASE, Const.CPU_LOWERCASE]:
            return npu_custom_functions[module.api_name](*args, **kwargs)
    return module.api_func(*args, **kwargs)


forward_methods = {
    "Tensor": tensor_module_forward,
    "Distributed": dist_module_forward,
    "NPU": npu_module_forward
}


class ApiTemplate(HOOKModule):
    def __init__(self, api_name, api_func, prefix, hook_build_func, need_hook=True, device=Const.CPU_LOWERCASE):
        self.api_name = api_name
        self.prefix = prefix
        self.prefix_api_name = prefix + Const.SEP + str(api_name.split(Const.SEP)[-1]) + Const.SEP
        self.need_hook = need_hook
        self.device = device
        self.op_is_distributed = prefix == Const.DIST_API_TYPE_PREFIX
        if self.need_hook:
            super().__init__(hook_build_func)
        self.api_func = api_func

    @torch_device_guard
    def forward(self, *args, **kwargs):
        exec_func = forward_methods.get(self.prefix)
        exec_func = functools.partial(exec_func, self) if exec_func else self.api_func
        return exec_func(*args, **kwargs)


api_register = None


def get_api_register(return_new=False):
    if return_new:
        return ApiRegistry(_api_types, _inner_used_api, _supported_api_list_path, ApiTemplate)

    global api_register
    if api_register is None:
        api_register = ApiRegistry(_api_types, _inner_used_api, _supported_api_list_path, ApiTemplate)
    return api_register



# -*- coding: utf-8 -*-
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import os
import sys
import numpy as np

from msprobe.core.common.log import logger
from msprobe.core.common.file_utils import change_mode
from msprobe.core.common.const import FileCheckConst
from msprobe.infer.offline.compare.msquickcmp.common.utils import execute_command
from msprobe.infer.utils.util import load_file_to_read_common_check, filter_cmd


MSACCUCMP_FILE_PATH = "toolkit/tools/operator_cmp/compare/msaccucmp.py"


def convert_bin_file_to_npy(bin_file_path, npy_dir_path, cann_path):
    """
    Function Description:
        convert a bin file to npy file.
    Parameter:
        bin_file_path: the path of the bin file needed to be converted to npy
        npy_dir_path: the dest dir to save the converted npy file
        cann_path: user or system cann_path for using msaccucmp.py
    """
    python_version = sys.executable.split('/')[-1]
    msaccucmp_command_file_path = os.path.join(cann_path, MSACCUCMP_FILE_PATH)
    logger.info(f"convert dump data: {bin_file_path} to npy file")
    bin2npy_cmd = [python_version, msaccucmp_command_file_path, "convert", "-d", bin_file_path, "-out", npy_dir_path]
    bin2npy_cmd = filter_cmd(bin2npy_cmd)
    execute_command(bin2npy_cmd)


def convert_npy_to_bin(npy_input_path):
    """
    Function Description:
        convert a  file to bin file.
    Parameter:
        npy_file_path: the path of the npy file needed to be converted to bin
    """
    input_initial_path = npy_input_path.split(",")
    outputs = []
    for input_item in input_initial_path:
        input_item_path = os.path.realpath(input_item)
        if input_item_path.endswith('.npy'):
            bin_item = input_item[:-4] + '.bin'
            bin_path = input_item_path[:-4] + '.bin'
            input_item_path = load_file_to_read_common_check(input_item_path)
            npy_data = np.load(input_item_path)
            if os.path.islink(bin_path):
                os.unlink(bin_path)
            if os.path.exists(bin_path):
                os.remove(bin_path)
            npy_data.tofile(bin_path)
            change_mode(bin_path, FileCheckConst.DATA_FILE_AUTHORITY)
            outputs.append(bin_item)
        else:
            outputs.append(input_item)
    return ",".join(outputs)



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import os
from msprobe.core.common.file_utils import load_yaml


class InplaceOpChecker:
    OP_FUNCTIONAL = 'functional'
    OP_TENSOR = 'tensor'
    OP_TORCH = 'torch'
    OP_DISTRIBUTED = 'distributed'

    INPLACE_OPS_DICT = None

    @classmethod
    def load_ops(cls):
        if cls.INPLACE_OPS_DICT is None:
            cls.INPLACE_OPS_DICT = dict()
        cur_path = os.path.dirname(os.path.realpath(__file__))
        yaml_path = os.path.join(cur_path, "inplace_ops.yaml")
        all_ops = load_yaml(yaml_path)
        cls.INPLACE_OPS_DICT[cls.OP_FUNCTIONAL] = all_ops.get('inplace_functional_op')
        cls.INPLACE_OPS_DICT[cls.OP_TENSOR] = all_ops.get('inplace_tensor_op')
        cls.INPLACE_OPS_DICT[cls.OP_TORCH] = all_ops.get('inplace_torch_op')
        cls.INPLACE_OPS_DICT[cls.OP_DISTRIBUTED] = all_ops.get('inplace_distributed_op')

    @classmethod
    def check(cls, api, category='distributed'):
        """
            给定api和分类，检查其是否为inplace操作
        """
        if not cls.INPLACE_OPS_DICT:
            cls.load_ops()

        if category not in cls.INPLACE_OPS_DICT.keys():
            return False
        return api in cls.INPLACE_OPS_DICT[category]


InplaceOpChecker.load_ops()



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import os
import re
from copy import deepcopy
from dataclasses import dataclass
from typing import ClassVar, Dict, List, Optional, Tuple

import yaml
from msprobe.core.common.const import Const
from msprobe.core.common.file_utils import save_yaml
from msprobe.core.common.log import logger
from msprobe.core.common.utils import CompareException, add_time_with_yaml
from msprobe.core.compare.layer_mapping.postprocess_pass import postprocess_pass


@dataclass
class DumpDataItem:
    framework: str
    data_name: Optional[str] = None
    api_type: Optional[str] = None
    api_name: Optional[str] = None
    type_name: Optional[str] = None
    full_scope: str = ""
    layer_scope: str = ""
    stack_scope: str = ""
    frame_stack_scope: str = ""
    user_stack_scope: str = ""
    construct_scope: str = ""
    scope_direction: Optional[str] = None
    scope_id: Optional[int] = None
    state: str = ""

    # 类变量使用 ClassVar
    layernames: ClassVar[set] = {Const.CELL, Const.MODULE}
    framework2stack_sign: ClassVar[Dict[str, Tuple[str, str]]] = {
        Const.MS_FRAMEWORK: ("Template", "construct"),
        Const.PT_FRAMEWORK: ("Template", r"in (for|back)ward,")
    }

    @staticmethod
    def check_stack_valid(stack_info):
        if stack_info is not None:
            if not isinstance(stack_info, list):
                logger.error(f"stack is invalid, it should be a list[str], but got {stack_info}")
                raise CompareException(CompareException.INVALID_DATA_ERROR)
            for stack in stack_info:
                if not isinstance(stack, str):
                    logger.error(f"stack is invalid, it should be a list[str], but got {stack_info}")
                    raise CompareException(CompareException.INVALID_DATA_ERROR)

    def set(self, data_name: str, construct_info: str, stack_info: str) -> None:
        self.set_name(data_name)
        self.set_layer_scope(construct_info)
        self.set_stack_scope(stack_info)
        self.set_full_scope()

    def set_name(self, data_name: str) -> None:
        self.data_name = data_name
        data_name_list = data_name.split(Const.SEP)
        if not data_name_list or len(data_name_list) < abs(Const.LAYER_NAME_INDEX):
            logger.error(
                f"The dump data does not comply with the format specification and "
                f"must contain no less than four fields. "
                f"The current data is {data_name}"
            )
            raise CompareException(CompareException.INVALID_DATA_ERROR)

        # 适配老的dump.json: xxx.word_embeddings.VocalParallelEmbedding.parameters_grad
        if data_name_list[Const.LAST_INDEX] == Const.PARAMS_GRAD:
            self.api_type = Const.PARAMS_GRAD
            self.api_name = data_name_list[Const.PARAMS_GRAD_NAME_INDEX]
            self.type_name = data_name_list[Const.PARAMS_GRAD_TYPE_NAME_INDEX]
            self.state = Const.PARAMS_GRAD
            return
        # 适配老的dump.json: xxx.word_embeddings.VocalParallelEmbedding.parameters_grad.数字
        if data_name_list[Const.LAST_INDEX - 1] == Const.PARAMS_GRAD and data_name_list[Const.LAST_INDEX].isdigit():
            self.api_type = Const.PARAMS_GRAD
            self.api_name = data_name_list[Const.PARAMS_GRAD_NAME_INDEX - 1]
            self.type_name = data_name_list[Const.PARAMS_GRAD_TYPE_NAME_INDEX - 1]
            self.state = Const.PARAMS_GRAD
            return

        self.api_type = data_name_list[Const.API_TYPE_INDEX]
        self.type_name = data_name_list[Const.TYPE_NAME_INDEX]
        if self.api_type in self.layernames:
            self.api_name = data_name_list[Const.LAYER_NAME_INDEX]
            self.state = data_name_list[Const.SCOPE_DIRECTION_INDEX]
        else:
            self.api_name = self.type_name
            self.state = data_name_list[Const.LAST_INDEX]

    def set_layer_scope(self, construct_info: str) -> None:
        self.construct_scope = construct_info
        if self.api_type in self.layernames:
            # remove api name
            data_list = self.data_name.split(Const.SEP)
            data_list = data_list[:Const.LAYER_NAME_INDEX] + data_list[Const.TYPE_NAME_INDEX:]
        elif self.api_type == Const.PARAMS_GRAD:
            data_list = self.data_name.split(Const.SEP)
        elif construct_info:
            data_list = construct_info.split(Const.SEP)
        else:
            data_list = []

        if data_list:
            self.layer_scope = Const.SEP.join(data_list[:Const.TYPE_NAME_INDEX])
        else:
            self.layer_scope = Const.TOP_LAYER
        if construct_info and Const.SEP in construct_info:
            construct_list = construct_info.split(Const.SEP)
            if len(construct_list) < abs(Const.LAYER_NAME_INDEX):
                logger.error(
                    f"The construct data does not comply with the format specification and "
                    f"must contain no less than four fields. "
                    f"The current data is {construct_info}"
                )
                raise CompareException(CompareException.INVALID_DATA_ERROR)
            self.scope_id = construct_list[Const.SCOPE_ID_INDEX]
            self.scope_direction = construct_list[Const.SCOPE_DIRECTION_INDEX]

    def set_stack_scope(self, stack_info: str) -> None:
        # Cell/Module has no stack info
        if self.api_type in self.layernames:
            return

        if self.api_type in Const.DATA_TYPE_SKIP_LIST or not stack_info:
            return

        start_sign, end_sign = self.framework2stack_sign.get(self.framework)
        self.check_stack_valid(stack_info)
        start_pos, end_pos = find_regard_scope(stack_info, start_sign, end_sign)
        # 获取指定范围的代码
        regard_scope = stack_info[start_pos + 1:end_pos]
        frame_func_stack_list, user_func_stack_list = find_stack_func_list(regard_scope)
        self.frame_stack_scope = Const.SEP.join(frame_func_stack_list)
        self.user_stack_scope = Const.SEP.join(user_func_stack_list)

    def set_full_scope(self, use_user_func_scope=False, use_frame_func_scope=True) -> None:
        scope_list = [self.layer_scope]
        if use_user_func_scope and self.user_stack_scope:
            scope_list.append(self.user_stack_scope)
        if use_frame_func_scope and self.frame_stack_scope:
            scope_list.append(self.frame_stack_scope)
        scope_list.append(self.api_name)
        self.full_scope = Const.SEP.join(scope_list)


def find_regard_scope(lines, start_sign, end_sign):
    # 找出 start_pos 和 end_pos
    start_pos = -1
    end_pos = len(lines)
    for idx, ii in enumerate(lines):
        if re.search(start_sign, ii):
            start_pos = idx
        elif start_pos >= 0 and re.search(end_sign, ii):
            end_pos = idx
            break
    return start_pos, end_pos


def find_stack_func_list(lines, record_user=True):
    res_list = []
    user_stack = []
    frame_stack = None
    no_entrance = True
    for line in lines:
        ele_list = line.split(Const.COMMA)
        file_ele = ele_list[Const.STACK_FILE_INDEX]
        # if framework func line and no framework entrance found yet
        if any(ii in file_ele for ii in Const.FRAME_FILE_LIST) and no_entrance:
            frame_stack = line  # Update the last target index
        else:
            if record_user:
                user_stack.append(line)
            no_entrance = False

    # Check if the last string in the list contains target str
    if frame_stack and no_entrance:
        no_entrance = False

    # 过滤和处理 regard_scope
    frame_func = get_stack_in_lines([frame_stack])
    user_func = get_stack_in_lines(user_stack)
    return (frame_func, user_func)


def get_stack_in_lines(simplified: List[str]):
    res_list = []
    if not simplified:
        return res_list
    for line in simplified:
        if not line:
            continue

        ele_list = line.split(Const.COMMA)
        file_ele = ele_list[Const.STACK_FILE_INDEX]
        if any(ii in file_ele for ii in Const.FILE_SKIP_LIST):
            continue

        func_ele = ele_list[Const.STACK_FUNC_INDEX]
        if any(ii in func_ele for ii in Const.FUNC_SKIP_LIST):
            continue

        in_func_name = func_ele.split()[Const.STACK_FUNC_ELE_INDEX]

        res_list.append(in_func_name)

    reversed_list = res_list[::-1]
    return reversed_list


def dumpdata_representer(dumper, data):
    d = deepcopy(data.__dict__)
    d.pop("data_name")
    return dumper.represent_dict(d)


def get_dump_data_items(dump, stack, construct, framework, output_path=None):
    if not stack or not construct:
        return []
    name2item = {}
    data_items = []

    dump_data = dump.get("data", {})
    for data_name in dump_data:
        code_info = stack.get(data_name, None)
        parent_info = construct.get(data_name, None)
        data_item = DumpDataItem(framework)
        data_item.set(data_name, parent_info, code_info)
        name2item[data_name] = data_item
        data_items.append(data_item)

    postprocess_pass(data_items, name2item)

    if output_path:
        yaml.add_representer(DumpDataItem, dumpdata_representer)
        file_name = add_time_with_yaml(f"{framework}_data")
        file_path = os.path.join(os.path.realpath(output_path), file_name)
        save_yaml(file_path, name2item)
    return data_items



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import re

from msprobe.core.common.const import Const
from msprobe.core.common.log import logger
from msprobe.core.common.exceptions import MsprobeException
from msprobe.core.common.utils import get_real_step_or_rank


class CommonConfig:
    def __init__(self, json_config):
        self.task = json_config.get('task')
        self.dump_path = json_config.get('dump_path')
        self.rank = get_real_step_or_rank(json_config.get('rank'), Const.RANK)
        self.step = get_real_step_or_rank(json_config.get('step'), Const.STEP)
        self.level = json_config.get('level')
        self.async_dump = json_config.get("async_dump", False)
        self.precision = json_config.get("precision", Const.DUMP_PRECISION_LOW)
        self._check_config()

    def _check_config(self):
        if self.task and self.task not in Const.TASK_LIST:
            logger.error_log_with_exp("task is invalid, it should be one of {}".format(Const.TASK_LIST),
                                      MsprobeException(MsprobeException.INVALID_PARAM_ERROR))
        if self.dump_path is not None and not isinstance(self.dump_path, str):
            logger.error_log_with_exp("dump_path is invalid, it should be a string",
                                      MsprobeException(MsprobeException.INVALID_PARAM_ERROR))
        if self.level and self.level not in Const.LEVEL_LIST:
            logger.error_log_with_exp("level is invalid, it should be one of {}".format(Const.LEVEL_LIST),
                                      MsprobeException(MsprobeException.INVALID_PARAM_ERROR))
        if not isinstance(self.async_dump, bool):
            logger.error_log_with_exp("async_dump is invalid, it should be a boolean",
                                      MsprobeException(MsprobeException.INVALID_PARAM_ERROR))
        elif self.async_dump:
            logger.warning("async_dump is True, it may cause OOM when dumping large tensor.")

        if self.precision not in Const.DUMP_PRECISION_LIST:
            logger.error_log_with_exp("precision is invalid, it should be one of {}".format(Const.DUMP_PRECISION_LIST),
                                      MsprobeException(MsprobeException.INVALID_PARAM_ERROR))


class BaseConfig:
    def __init__(self, json_config):
        self.scope = json_config.get('scope')
        self.list = json_config.get('list')
        self.data_mode = json_config.get('data_mode')
        self.file_format = json_config.get("file_format")
        self.summary_mode = json_config.get("summary_mode")
        self.diff_nums = json_config.get("diff_nums")
        self.check_mode = json_config.get("check_mode")
        self.is_regex_valid = True

    @staticmethod
    def _check_str_list_config(config_item, config_name):
        if config_item is not None:
            if not isinstance(config_item, list):
                logger.error_log_with_exp(f"{config_name} is invalid, it should be a list[str]",
                                          MsprobeException(MsprobeException.INVALID_PARAM_ERROR))
            for name in config_item:
                if not isinstance(name, str):
                    logger.error_log_with_exp(f"{config_name} is invalid, it should be a list[str]",
                                              MsprobeException(MsprobeException.INVALID_PARAM_ERROR))

    def check_config(self):
        self._check_str_list_config(self.scope, "scope")
        self._check_str_list_config(self.list, "list")
        self._check_data_mode()
        self._check_regex_in_list()

    def _check_data_mode(self):
        if self.data_mode is not None:
            if not isinstance(self.data_mode, list):
                logger.error_log_with_exp("data_mode is invalid, it should be a list[str]",
                                          MsprobeException(MsprobeException.INVALID_PARAM_ERROR))

            if Const.ALL in self.data_mode and len(self.data_mode) != 1:
                logger.error_log_with_exp(
                    "'all' cannot be combined with other options in data_mode.",
                    MsprobeException(MsprobeException.INVALID_PARAM_ERROR)
                )

            if len(self.data_mode) >= len(Const.DUMP_DATA_MODE_LIST):
                logger.error_log_with_exp(
                    f"The number of elements in the data_made cannot exceed {len(Const.DUMP_DATA_MODE_LIST) - 1}.",
                    MsprobeException(MsprobeException.INVALID_PARAM_ERROR)
                )

            for mode in self.data_mode:
                if not isinstance(mode, str):
                    logger.error_log_with_exp("data_mode is invalid, it should be a list[str]",
                                              MsprobeException(MsprobeException.INVALID_PARAM_ERROR))
                if mode not in Const.DUMP_DATA_MODE_LIST:
                    logger.error_log_with_exp(
                        f"The element '{mode}' of data_mode {self.data_mode} is not in {Const.DUMP_DATA_MODE_LIST}.",
                        MsprobeException(MsprobeException.INVALID_PARAM_ERROR)
                    )

    def _check_summary_mode(self):
        if self.summary_mode and self.summary_mode not in Const.SUMMARY_MODE:
            logger.error_log_with_exp(
                        f"summary_mode is invalid, summary_mode is not in {Const.SUMMARY_MODE}.",
                        MsprobeException(MsprobeException.INVALID_PARAM_ERROR)
                    )

    def _check_regex_in_list(self):
        if self.list:
            for name in self.list:
                if name.startswith('name-regex(') and name.endswith(')'):
                    try:
                        re.compile(name[len('name-regex('):-1])
                    except re.error:
                        self.is_regex_valid = False
                        break



# coding=utf-8
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

"""
Function:
This file mainly involves the dump function.
"""
import re
import os
from enum import Enum
from enum import unique

from msprobe.msaccucmp.cmp_utils import utils, utils_type, path_check
from msprobe.msaccucmp.cmp_utils import log
from msprobe.msaccucmp.cmp_utils.constant.const_manager import ConstManager
from msprobe.msaccucmp.cmp_utils.reg_manager import RegManager
from msprobe.msaccucmp.cmp_utils.constant.compare_error import CompareError
from msprobe.msaccucmp.dump_parse.ffts_parser import FFTSParser
from msprobe.msaccucmp.dump_parse import dump_utils, mapping
from msprobe.msaccucmp.dump_parse.dump_data_object import DumpDataObj, DumpTensor


@unique
class DumpType(Enum):
    """
    The enum for dump type
    """
    Offline = 0
    Standard = 1
    Quant = 2
    Numpy = 3


class DumpInfo:
    """
    The class for dump info
    """

    def __init__(self: any, dump_path: str, dump_version: int, ffts: bool = False,
                 fusion_json_file_path: str = "") -> None:
        self.path = dump_path
        self.type = None
        self.op_name_to_file_map = {}
        self.op_name_to_task_mode_map = {}
        self.quant = False
        self.data_info = ''
        self.dump_version = dump_version
        self.hash_to_file_name_map = {}
        self.ffts = ffts
        self.fusion_json_file_path = fusion_json_file_path

    def check_arguments_valid(self: any) -> None:
        """
        Check arguments valid, if invalid, throw exception
        """
        ret = path_check.check_path_valid(self.path, True, False, path_check.PathType.Directory)
        if ret != CompareError.MSACCUCMP_NONE_ERROR:
            raise CompareError(ret)
        self._make_op_name_to_file_map()

    def is_standard_quant(self: any) -> bool:
        """
        The dump data is standard quant
        :return: bool, true if dump data is standard quant
        """
        return self.type == DumpType.Quant or (self.type == DumpType.Numpy and self.quant)

    def is_standard_origin(self: any) -> bool:
        """
        The dump data is standard origin
        :return: bool, true if dump data is standard origin
        """
        return self.type == DumpType.Standard or (self.type == DumpType.Numpy and not self.quant)

    def get_op_dump_file(self: any, op_name: str, output_index: int = None, print_log: bool = True) -> tuple:
        """
        Get the dump file a by op name
        :param op_name: the op name
        :param output_index: the output index
        :param print_log: print log or not
        :return: the dump file path
        """
        original_op_name = op_name.replace('/', '_').replace('.', '_')
        if output_index is not None:
            original_op_name = '%s.%d' % (original_op_name, output_index)
        if original_op_name not in self.op_name_to_file_map:
            raise CompareError(CompareError.MSACCUCMP_NO_DUMP_FILE_ERROR)
        dump_file_list = self.op_name_to_file_map.get(original_op_name)
        dump_mode = self.op_name_to_task_mode_map.get(original_op_name)
        if not dump_file_list:
            raise CompareError(CompareError.MSACCUCMP_NO_DUMP_FILE_ERROR)
        if len(dump_file_list) > 1:
            dump_file_list = dump_utils.sort_dump_file_list(dump_mode, dump_file_list)
        if print_log:
            log.print_info_log('[%s] [%s] %s' % (op_name, str(self.type.name), dump_file_list))
        return dump_file_list, dump_mode

    def get_op_dump_data(self: any, op_name: str, output_index: int = None) -> (str, DumpDataObj):
        """
        Get the dump file and data by op name
        :param op_name: the op name
        :param output_index: the output index
        :return: the dump file path, dump data
        """
        if self.type == DumpType.Quant and output_index is None:
            output_index = 0
        dump_file_list, dump_mode = self.get_op_dump_file(op_name, output_index)
        dump_data_list = [dump_utils.parse_dump_file(dump_file_path,
                                                     self.dump_version) for dump_file_path in dump_file_list]
        if dump_mode == ConstManager.AUTOMATIC_MODE or dump_mode == ConstManager.MANUAL_MODE:
            ffts_parser = FFTSParser(dump_file_list, dump_data_list)
            dump_file_path, dump_data = ffts_parser.parse_ffts
        else:
            dump_file_path = dump_file_list[-1]
            dump_data = dump_data_list[-1]
        return dump_file_path, dump_data

    def get_data_info(self: any) -> str:
        """
        Get the dump data info
        :return: dump data info
        """
        if self.data_info == '':
            self.data_info = 'side is dump data of the '
            prefix = 'quantized' if self.quant else 'unquantized'
            if self.type == DumpType.Offline:
                self.data_info += prefix + ' model executed on the AI processor'
            elif self.type == DumpType.Standard:
                self.data_info += 'unquantized original model'
            elif self.type == DumpType.Quant:
                self.data_info += 'quantized original model'
        return self.data_info

    def _match_dump_pattern(self: any, pattern: str, item: str, info: str, expect: DumpType) -> (any, DumpType):
        _, match = RegManager.match_group(pattern, item)
        if match is None:
            if item in self.hash_to_file_name_map.values():
                msg = 'The file name \"{}\" in the file \"{}\" is invalid. It only supports {}' \
                    .format(item, os.path.join(self.path, ConstManager.MAPPING_FILE_NAME), info)
            else:
                msg = 'The file name \"{}\" in the path \"{}\"is invalid, only supports {}' \
                    .format(item, self.path, info)
            log.print_warn_log(msg)
            raise CompareError(CompareError.MSACCUCMP_DUMP_FILE_ERROR)
        return match.group(1), expect

    def _check_dump_file_is_quant(self: any, dump_type: DumpType, op_name: str) -> None:
        if dump_type in [DumpType.Offline, DumpType.Numpy]:
            suffix_list = ConstManager.QUANT_OP_NANE_SUFFIX_LIST + ConstManager.DEQUANT_OP_NANE_SUFFIX_LIST
            for suffix in suffix_list:
                if suffix in op_name:
                    self.quant = True

    def _check_file_match_pattern(self: any, item: str) -> (any, DumpType):
        if item.endswith(ConstManager.NUMPY_SUFFIX):
            return self._match_dump_pattern(
                RegManager.NUMPY_DUMP_PATTERN, item, ConstManager.NUMPY_FILE_NAME, DumpType.Numpy)
        if item.endswith(ConstManager.STANDARD_SUFFIX):
            return self._match_dump_pattern(
                RegManager.STANDARD_DUMP_PATTERN, item, ConstManager.STANDARD_FILE_NAME, DumpType.Standard)
        if item.endswith(ConstManager.QUANT_SUFFIX):
            return self._match_dump_pattern(
                RegManager.QUANT_DUMP_PATTERN, item, ConstManager.QUANT_FILE_NAME, DumpType.Quant)
        return self._match_dump_pattern(
            RegManager.OFFLINE_DUMP_PATTERN, item, ConstManager.OFFLINE_FILE_NAME, DumpType.Offline)

    def _handle_one_file(self: any, file_path: str) -> None:
        item = os.path.basename(file_path)
        item_name, item_extension = os.path.splitext(item)
        if item.isdigit() or (item_name.isdigit() and item_extension == ".npy"):
            mapping_file_path = os.path.join(self.path, ConstManager.MAPPING_FILE_NAME)
            if not os.path.exists(mapping_file_path):
                log.print_warn_log('The file name \"{}\" corresponding mapping file \"{}\" is not exist.'
                                   .format(item, mapping_file_path))
                return
            if self.hash_to_file_name_map and not self.hash_to_file_name_map.get(item):
                log.print_warn_log('The file name \"{}\" in the file \"{}\" is not exist.'
                                   .format(item, mapping_file_path))
                return
            item = self.hash_to_file_name_map.get(item)
        try:
            match, current_dump_type = self._check_file_match_pattern(item)
        except CompareError:
            return
        finally:
            pass
        if not self.ffts:
            op_name = handle_op_name(match, self.fusion_json_file_path)
        # if real op name contain '_lxslice' field, the op will not be added to map
            if ConstManager.FFTS_MANUAL_MODE_FIELD in op_name:
                return
            self._check_task_type(op_name, item)
        else:
            op_name = match

        self._check_task_type(op_name, item)
        self._check_dump_file_is_quant(current_dump_type, op_name)
        if self.type is None:
            self.type = current_dump_type
        else:
            if self.type != current_dump_type:
                log.print_error_log('Not all files in the path "%r" are of the same type, such as "%s".'
                                    % (self.path, item))
                raise CompareError(CompareError.MSACCUCMP_DUMP_FILE_ERROR)
        if op_name in self.op_name_to_file_map:
            self.op_name_to_file_map.get(op_name).append(file_path)
        else:
            self.op_name_to_file_map[op_name] = [file_path]

    def _make_op_name_to_file_map(self: any) -> None:
        mapping_file_path = os.path.join(self.path, ConstManager.MAPPING_FILE_NAME)
        self.hash_to_file_name_map = mapping.read_mapping_file(mapping_file_path)
        for item in os.listdir(self.path):
            if item in (ConstManager.MAPPING_FILE_NAME, ConstManager.CONVERT_FAILED_FILE_LIST_NAME):
                continue
            file_path = os.path.join(self.path, item)
            if os.path.isfile(file_path):
                self._handle_one_file(file_path)
        self._judge_dump_type()

    def _check_task_type(self: any, op_name: str, file_name: str) -> None:
        if file_name.endswith(ConstManager.NUMPY_SUFFIX):
            self.op_name_to_task_mode_map[op_name] = ConstManager.NORMAL_MODE
            return
        flied_list = file_name.split(".")
        if len(flied_list) <= ConstManager.OLD_FILE_FIELD_NUM + 1:
            self.op_name_to_task_mode_map[op_name] = ConstManager.NORMAL_MODE
            return
        if flied_list[-4] not in ConstManager.TASK_TYPE_MAP.values():
            raise CompareError(CompareError.MSACCUCMP_INVALID_TASK_TYPE)
        if flied_list[-4] == ConstManager.TASK_TYPE_MAP.get(ConstManager.FFTSPLUS):
            if ConstManager.FFTS_MANUAL_MODE_FIELD in file_name:
                self.op_name_to_task_mode_map[op_name] = ConstManager.MANUAL_MODE
            else:
                self.op_name_to_task_mode_map[op_name] = ConstManager.AUTOMATIC_MODE
            return
        self.op_name_to_task_mode_map[op_name] = ConstManager.NORMAL_MODE

    def _judge_dump_type(self: any) -> None:
        if self.type is None:
            log.print_error_log('No valid dump files in the path "%r".' % self.path)
            raise CompareError(CompareError.MSACCUCMP_DUMP_FILE_ERROR)
        if self.type == DumpType.Numpy:
            if self.quant:
                self.type = DumpType.Quant
            else:
                self.type = DumpType.Standard


class CompareData:
    """
    The class for compare data, left dump data and right dump data
    """

    def __init__(self: any, left_dump_path: str, right_dump_path: str, dump_version: int,
                 ffts: bool = False, fusion_json_file_path: str = "") -> None:
        self.left_dump_info = DumpInfo(left_dump_path, dump_version, ffts, fusion_json_file_path)
        self.right_dump_info = DumpInfo(right_dump_path, dump_version, ffts, fusion_json_file_path)
        self.dump_version = dump_version

    def check_arguments_valid(self: any, fusion_json_file_path: str,
                              quant_fusion_rule_file_path: str,
                              close_fusion_rule_file_path: str) -> None:
        """
        Check arguments valid, if invalid, throw exception
        :param fusion_json_file_path: the fusion json file path
        :param quant_fusion_rule_file_path: the quant fusion rule file path
        :param close_fusion_rule_file_path: the clsoe fusion json file path
        """
        self.left_dump_info.check_arguments_valid()
        self.right_dump_info.check_arguments_valid()
        left_type = self.left_dump_info.type
        info = 'When the left %s and the right %s,' \
               % (self.left_dump_info.get_data_info(), self.right_dump_info.get_data_info())
        if left_type == DumpType.Offline:
            self._check_left_type_offline_valid(info, fusion_json_file_path, quant_fusion_rule_file_path,
                                                close_fusion_rule_file_path)
        elif left_type == DumpType.Quant:
            self._check_left_type_quant_valid(info, fusion_json_file_path, quant_fusion_rule_file_path,
                                              close_fusion_rule_file_path)
        elif left_type == DumpType.Standard:
            log.print_error_log('%s this scenario cannot be compared.' % info)
            raise CompareError(CompareError.MSACCUCMP_INVALID_DUMP_TYPE_ERROR)

    def get_left_dump_data(self: any, op_name: str) -> (str, DumpDataObj):
        """
        Get the left dump file path and data by fusion op
        :param op_name: the op name
        :return: the left dump file path, the left dump data
        """
        return self.left_dump_info.get_op_dump_data(op_name)

    def get_right_dump_data(self: any, op_name: str, output_index: int = 0) -> (str, DumpDataObj):
        """
        Get the right dump file path and data by fusion op
        :param op_name: the op name
        :param output_index: the output index
        :return: the right dump file path, the right dump data
        """
        if self.right_dump_info.type == DumpType.Offline:
            return self.right_dump_info.get_op_dump_data(op_name)
        return self.right_dump_info.get_op_dump_data(op_name, output_index)

    def is_standard_quant_vs_origin(self: any) -> bool:
        """
        The scenario is quantized original vs unquantized original
        :return: bool, true if the scenario is quantized original vs unquantized original
        """
        return self.left_dump_info.is_standard_quant() and self.right_dump_info.is_standard_origin()

    def _check_offline_standard_valid(self: any, info: str, fusion_json_file_path: str,
                                      quant_fusion_rule_file_path: str, close_fusion_rule_file_path: str) -> None:
        if fusion_json_file_path == "":
            log.print_error_log('%s the -f parameter is required.' % info)
            raise CompareError(CompareError.MSACCUCMP_INVALID_PARAM_ERROR)
        if close_fusion_rule_file_path != '':
            log.print_error_log('%s there is no need to enter the -cf parameter.' % info)
            raise CompareError(CompareError.MSACCUCMP_INVALID_PARAM_ERROR)
        if self.left_dump_info.quant and quant_fusion_rule_file_path == '':
            log.print_error_log('%s the -q parameter is required.' % info)
            raise CompareError(CompareError.MSACCUCMP_INVALID_PARAM_ERROR)
        if not self.left_dump_info.quant and quant_fusion_rule_file_path != '':
            log.print_error_log('%s there is no need to enter the -q parameter.' % info)
            raise CompareError(CompareError.MSACCUCMP_INVALID_PARAM_ERROR)

    def _check_offline_quant_valid(self: any, info: str, fusion_json_file_path: str,
                                   quant_fusion_rule_file_path: str, close_fusion_rule_file_path: str) -> None:
        if not self.left_dump_info.quant:
            log.print_error_log('%s this scenario cannot be compared.' % info)
            raise CompareError(CompareError.MSACCUCMP_INVALID_DUMP_TYPE_ERROR)
        if fusion_json_file_path == "":
            log.print_error_log('%s the -f parameter is required.' % info)
            raise CompareError(CompareError.MSACCUCMP_INVALID_PARAM_ERROR)
        if quant_fusion_rule_file_path != '' or close_fusion_rule_file_path != '':
            log.print_error_log('%s there is no need to enter the -q or -cf parameter.' % info)
            raise CompareError(CompareError.MSACCUCMP_INVALID_PARAM_ERROR)

    def _check_left_type_offline_valid(self: any, info: str, fusion_json_file_path: str,
                                       quant_fusion_rule_file_path: str, close_fusion_rule_file_path: str) -> None:
        right_type = self.right_dump_info.type
        if right_type == DumpType.Offline:
            if self.left_dump_info.quant != self.right_dump_info.quant:
                log.print_error_log('%s this scenario cannot be compared.' % info)
                raise CompareError(CompareError.MSACCUCMP_INVALID_DUMP_TYPE_ERROR)
            if fusion_json_file_path != '' and close_fusion_rule_file_path == '':
                log.print_error_log('%s the -cf parameter is required.' % info)
                raise CompareError(CompareError.MSACCUCMP_INVALID_PARAM_ERROR)
            if fusion_json_file_path == '' and close_fusion_rule_file_path != '':
                log.print_error_log('%s the -f parameter is required.' % info)
                raise CompareError(CompareError.MSACCUCMP_INVALID_PARAM_ERROR)
            if quant_fusion_rule_file_path != '':
                log.print_error_log('%s there is no need to enter the -q parameter.' % info)
                raise CompareError(CompareError.MSACCUCMP_INVALID_PARAM_ERROR)
        elif right_type == DumpType.Standard:
            self._check_offline_standard_valid(info, fusion_json_file_path, quant_fusion_rule_file_path,
                                               close_fusion_rule_file_path)
        elif right_type == DumpType.Quant:
            self._check_offline_quant_valid(info, fusion_json_file_path, quant_fusion_rule_file_path,
                                            close_fusion_rule_file_path)

    def _check_left_type_quant_valid(self: any, info: str, fusion_json_file_path: str,
                                     quant_fusion_rule_file_path: str, close_fusion_rule_file_path: str) -> None:
        right_type = self.right_dump_info.type
        if close_fusion_rule_file_path != "":
            log.print_error_log('%s there is no need to enter the -cf parameter.' % info)
            raise CompareError(CompareError.MSACCUCMP_INVALID_DUMP_TYPE_ERROR)
        if right_type in (DumpType.Offline, DumpType.Quant):
            log.print_error_log('%s this scenario cannot be compared.' % info)
            raise CompareError(CompareError.MSACCUCMP_INVALID_DUMP_TYPE_ERROR)
        elif right_type == DumpType.Standard:
            if quant_fusion_rule_file_path == '':
                log.print_error_log('%s the -q parameter is required.' % info)
                raise CompareError(CompareError.MSACCUCMP_INVALID_DUMP_TYPE_ERROR)
            if fusion_json_file_path != "":
                log.print_error_log('%s there is no need to enter the -f parameter.' % info)
                raise CompareError(CompareError.MSACCUCMP_INVALID_DUMP_TYPE_ERROR)


def handle_op_name(file_op_name: str, fusion_json_file_path: str) -> str:
    if fusion_json_file_path and ConstManager.FFTS_MANUAL_MODE_FIELD in file_op_name:
        file_op_name = process_op_name(file_op_name)
        return file_op_name

    # filter field '_lxsliceX' and '_sgt_field'
    if ConstManager.FFTS_MANUAL_MODE_FIELD not in file_op_name \
            and ConstManager.SGT_FIELD not in file_op_name:
        return file_op_name

    # field '_lxsliceX' at the end of name
    if ConstManager.FFTS_MANUAL_MODE_FIELD in file_op_name:
        first_match = RegManager.get_matchs(
            RegManager.FFTS_MANUAL_FIELD_PATTERN, file_op_name)[0]
        file_op_name = \
            file_op_name[:first_match.start() - 1] if first_match.end() == first_match.endpos else file_op_name

    # filter field '_sgt_field'
    if ConstManager.SGT_FIELD in file_op_name:
        # field '_sgt_graph' in the name
        end_match = RegManager.get_matchs(
            RegManager.SGT_FLIED_PATTERN, file_op_name)[-1]
        file_op_name = file_op_name[end_match.end() + 1:] if end_match.end() != end_match.endpos else file_op_name
    return file_op_name


def process_op_name(name):
    re_pattern = re.compile(RegManager.LXSLICE_PATTERN)
    op_name = re_pattern.sub("", name)
    return op_name




# coding=utf-8
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

"""
Function:
ShapeConversion class. This class mainly involves the convert_shape function.
"""
import sys
import time
from msprobe.msaccucmp.cmp_utils import log, file_utils
from msprobe.msaccucmp.conversion.shape_format_conversion import ShapeConversionMain


if __name__ == "__main__":
    log.print_deprecated_warning(sys.argv[0])
    START = time.time()
    SHAPE_CONVERSION = ShapeConversionMain()
    RET = 0
    with file_utils.UmaskWrapper():
        try:
            RET = SHAPE_CONVERSION.process()
        except Exception as base_err:
            log.print_error_log(f'Basic error running {sys.argv[0]}: {base_err}')
            sys.exit(1)
    END = time.time()
    log.print_info_log("The format conversion was completed and took %.2f seconds." % (END - START))
    sys.exit(RET)



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

from dataclasses import dataclass

from typing import Dict, List

from msprobe.core.common.const import Const


@dataclass
class APIInfo:
    api_name: str
    torch_api_name: str
    input_args: List[Dict]
    input_kwargs: Dict
    output_data: List[Dict]

    def __init__(self, api_name, input_args=None, input_kwargs=None, output_data=None):
        self.api_name = api_name
        self.input_args = input_args
        self.input_kwargs = input_kwargs
        self.output_data = output_data
        self.torch_api_name = self.extract_torch_api(self.api_name)

    @staticmethod
    def extract_torch_api(api_name) -> str:
        """
        Process tensor api name to extract first two fields in lowercase.
        """
        # Empty string checking
        if not api_name.strip():
            return ""

        parts = api_name.split(Const.SEP)

        # Handle different cases based on number of parts
        if len(parts) == 0:
            return ""
        elif len(parts) == 1:
            return parts[0].lower()
        else:
            return Const.SEP.join(parts[:2]).lower()



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import os.path
from dataclasses import dataclass, field
from typing import Set

from msprobe.core.common.file_utils import load_yaml
from msprobe.core.overflow_check.api_info import APIInfo
from msprobe.core.overflow_check.utils import has_nan_inf

cur_path = os.path.dirname(os.path.realpath(__file__))


class IgnoreFilter:
    def __init__(self, rule_path=os.path.join(cur_path, './ignore_rules.yaml')):
        self.rules = dict()
        self._load_rules(rule_path)

    def has_api_rule(self, api_name: str) -> bool:
        return api_name in self.rules.keys()

    def apply_filter(self, api_info: APIInfo) -> bool:
        """
            应用过滤规则，返回是否需要被过滤
        Args:
            api_info: API调用信息
        Returns:
            是否为误检，是否需要过滤
        """
        torch_api = api_info.torch_api_name
        if not self.has_api_rule(torch_api):
            return False
        rule = self.rules.get(torch_api)
        if not rule.match(api_info):
            return False
        return True

    def _load_rules(self, rule_file_path):
        if self.rules and len(self.rules):
            return
        data = load_yaml(rule_file_path)
        self.rules = dict()
        for rule_item in data.get('ignore_nan_inf', []):
            rule = Rule(
                api_name=rule_item.get('api_name', ''),
                desc=rule_item.get('description', ''),
                input_ignore=rule_item.get('input_ignore', []),
                output_ignore=rule_item.get('output_ignore', [])
            )
            if not rule.verify_field():
                continue
            if self.has_api_rule(rule.api_name):
                continue
            self.rules[rule.api_name] = rule


class Rule:

    def __init__(self, api_name, desc='', input_ignore=None, output_ignore=None):
        self.api_name = api_name
        self.desc = desc
        self.input_ignore = IgnoreItem()
        self.output_ignore = IgnoreItem()
        self._init_ignore(input_ignore, output_ignore)

    def __repr__(self):
        return (f'Rule(api_name={self.api_name}, desc={self.desc}, input_ignore={self.input_ignore}, output_ignore='
                f'{self.output_ignore})')

    def verify_field(self):
        if self.api_name == '':
            return False
        # 若无输入输出规则长度，则为无效规则
        if not (len(self.input_ignore.index) + len(self.input_ignore.name) + len(self.output_ignore.index)):
            return False
        return True

    def match(self, api_info: APIInfo) -> bool:
        """
            匹配API信息是否符合规则
        Returns:
            bool: True if the api_info matches this rule, False otherwise
        """
        # 首先检查API名称是否匹配
        api_name = api_info.torch_api_name
        if api_name != self.api_name:
            return False

        # 检查输入参数中的NaN/Inf
        if self.input_ignore.index and len(api_info.input_args):
            for idx, arg in enumerate(api_info.input_args):
                if has_nan_inf(arg) and not self.input_ignore.has_index(idx):
                    return False

        # 检查输入kwargs中的NaN/Inf
        if self.input_ignore.name and len(api_info.input_kwargs):
            for name, value in api_info.input_kwargs.items():
                if has_nan_inf(value) and not self.input_ignore.has_name(name):
                    return False

        # 检查输出中的NaN/Inf
        if self.output_ignore.index and len(api_info.output_data):
            for idx, out in enumerate(api_info.output_data):
                if has_nan_inf(out) and not self.output_ignore.has_index(idx):
                    return False

        return True

    def _init_ignore(self, input_ignore=None, output_ignore=None):
        """初始化忽略项"""
        if input_ignore is None:
            input_ignore = []
        if output_ignore is None:
            output_ignore = []

        # 处理输入忽略规则
        for item in input_ignore:
            if 'index' in item:
                self.input_ignore.add_index(item['index'])
            if 'name' in item:
                self.input_ignore.add_name(item['name'])

        # 处理输出忽略规则
        for item in output_ignore:
            if 'index' in item:
                self.output_ignore.add_index(item['index'])


@dataclass
class IgnoreItem:
    """存储需要忽略的索引和名称"""
    index: Set[int] = field(default_factory=set)
    name: Set[str] = field(default_factory=set)

    def add_index(self, idx: int):
        self.index.add(idx)

    def add_name(self, name: str):
        self.name.add(name)

    def has_index(self, idx: int) -> bool:
        return idx in self.index

    def has_name(self, name: str) -> bool:
        return name in self.name



# -*- coding: utf-8 -*-
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

from enum import Enum


class Arg(object):
    def __init__(self, benchmark_arg, atc_arg, msquickcmp_arg):
        self.atc_arg = atc_arg
        self.benchmark_arg = benchmark_arg
        self.msquickcmp_arg = msquickcmp_arg


class DynamicArgumentEnum(Enum):
    # enum struct Arg(benchmark_arg, atc_arg, msquickcmp_arg)
    DYM_BATCH = Arg("--dymBatch", "--dynamic_batch_size", None)
    DYM_SHAPE = Arg("--dymShape", "--input_shape_range", "input_shape")
    DYM_DIMS = Arg("--dymDims", "--dynamic_dims", "input_shape")

    @staticmethod
    def get_all_args() -> list:
        """
        get all argument enum, return as a list
        """
        return list(map(lambda arg: arg, DynamicArgumentEnum))



# coding=utf-8
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

"""
Function:
DetailComparison class. This class mainly involves the compare function.
"""

import os

from msprobe.msaccucmp.vector_cmp.fusion_manager import fusion_rule_parser
from msprobe.msaccucmp.cmp_utils import utils, utils_type, path_check
from msprobe.msaccucmp.cmp_utils import log
from msprobe.msaccucmp.cmp_utils.constant.compare_error import CompareError
from msprobe.msaccucmp.cmp_utils.constant.const_manager import ConstManager
from msprobe.msaccucmp.dump_parse import dump, mapping
from msprobe.msaccucmp.vector_cmp.compare_detail.detail_writer import DetailWriter
from msprobe.msaccucmp.vector_cmp.compare_detail.detail import DetailInfo
from msprobe.msaccucmp.vector_cmp.fusion_manager.compare_fusion_op import FusionOpComparison
from msprobe.msaccucmp.conversion.tensor_conversion import TensorConversion


class DetailComparison:
    """
    The class for fusion op compare
    """

    def __init__(self: any, detail_info: DetailInfo, fusion_op_comparison: FusionOpComparison,
                 output_path: str) -> None:
        self.detail_info = detail_info
        self.fusion_op_comparison = fusion_op_comparison
        self.detail_writer = DetailWriter(output_path, detail_info)
        self.fusion_op = None

    def check_index_valid(self: any, my_output_data: any) -> None:
        """
        Check index valid
        :param my_output_data: my output data
        """
        if self.detail_info.tensor_id.is_input():
            count = len(my_output_data.input_data)
        else:
            count = len(my_output_data.output_data)
        if self.detail_info.tensor_id.index >= count:
            log.print_out_of_range_error(self.fusion_op.op_name, self.detail_info.tensor_id.tensor_type,
                                         self.detail_info.tensor_id.index, '[0, %d)' % count)
            raise CompareError(CompareError.MSACCUCMP_INDEX_OUT_OF_BOUNDS_ERROR)

    def compare(self: any) -> int:
        """
        Compare detail by op name
        :return error_code
        """
        tensor_id = self.detail_info.tensor_id.get_tensor_id()
        log.print_info_log('[%s] Start to compare detail for %s.'
                           % (self.detail_info.tensor_id.op_name, tensor_id))
        # get fusion op list by op name
        tensor_list, dump_file_name = self._get_my_output_tensor_list()

        # delete old result
        try:
            self.detail_writer.delete_old_detail_result_files()
        except (OSError, SystemError, ValueError, TypeError, RuntimeError,
                MemoryError, KeyError, IOError) as error:
            log.print_error_log('Failed to delete the old detail result file. %s' % error)
            raise CompareError(CompareError.MSACCUCMP_DELETE_FILE_ERROR) from error

        # get right dump data by original op name and index
        try:
            ground_truth_tensor = \
                self.fusion_op_comparison.get_right_dump_data(
                    self.fusion_op, self.detail_info.tensor_id.index, self.detail_info.tensor_id.is_input())
        except CompareError as compare_error:
            if compare_error.code == CompareError.MSACCUCMP_NO_DUMP_FILE_ERROR:
                log.print_no_right_dump_file_error(self.fusion_op.op_name, tensor_id, is_error=True)
            raise compare_error

        # get format by right output info
        self.detail_info.set_detail_format(utils.convert_shape_to_string(
            tensor_list[self.detail_info.tensor_id.index].shape),
            tensor_list[self.detail_info.tensor_id.index].tensor_format,
            ground_truth_tensor.tensor_format
        )
        # deserialize output data to array
        tensor_conversion = TensorConversion(self.fusion_op, self.fusion_op_comparison.format_manager,
                                             is_detail=True)
        my_output_array, ground_truth_array, my_output_shape = \
            tensor_conversion.get_my_output_and_ground_truth_data(
                self.fusion_op_comparison.compare_data, tensor_list[self.detail_info.tensor_id.index],
                ground_truth_tensor)
        self.detail_writer.write(my_output_shape, my_output_array, ground_truth_array, dump_file_name)
        return CompareError.MSACCUCMP_NONE_ERROR

    def _print_l1_fusion_warning(self: any) -> None:
        timestamp_list = self.fusion_op_comparison.sort_l1_fusion_dump_file()
        max_timestamp_op_name = timestamp_list[-1][ConstManager.FUSION_OP_INDEX].op_name
        if self.fusion_op.op_name != max_timestamp_op_name:
            log.print_warn_log(
                'The dump data of %s is incomplete, the comparison may be far away. '
                'The dump data of %s may be complete, It is best to use %s for comparison.'
                % (self.fusion_op.op_name, max_timestamp_op_name, max_timestamp_op_name))

    def _get_my_output_tensor_by_op(self: any, relation: int) -> any:
        my_output_data_path, my_output_data = \
            self.fusion_op_comparison.compare_data.get_left_dump_data(self.fusion_op.op_name)
        dump_file = os.path.basename(my_output_data_path)
        self.check_index_valid(my_output_data)
        if self.detail_info.tensor_id.is_input():
            tensor_list = my_output_data.input_data
        else:
            if relation == utils_type.FusionRelation.L1Fusion:
                self._print_l1_fusion_warning()
            tensor_list = my_output_data.output_data
        return tensor_list, dump_file

    def _get_my_output_tensor_list(self: any) -> list:
        self.fusion_op, fusion_op_list = self.detail_info.get_detail_op(
            self.fusion_op_comparison.compare_rule.fusion_info)
        if self.fusion_op.is_inner_node():
            log.print_skip_inner_op_msg(self.fusion_op.op_name, is_error=True)
            raise CompareError(CompareError.MSACCUCMP_UNSUPPORTED_COMPARE_ERROR)
        if self.fusion_op.attr.quant_filter:
            log.print_error_log("[{}] The -op param should not specify an operator in quant/dequant pair."
                                .format(self.fusion_op.op_name))
            raise CompareError(CompareError.MSACCUCMP_UNSUPPORTED_COMPARE_ERROR)
        relation = fusion_rule_parser.get_relation_for_fusion(fusion_op_list)
        try:
            tensor_list, dumpfile = self._get_my_output_tensor_by_op(relation)
        except CompareError as compare_error:
            if compare_error.code == CompareError.MSACCUCMP_NO_DUMP_FILE_ERROR:
                log.print_no_left_dump_file_error(self.fusion_op.op_name, self.fusion_op.op_type, True)
            raise compare_error
        finally:
            pass
        return tensor_list, dumpfile


class DumpDetailComparison:
    """
    The class for Npu dump op compare
    """

    def __init__(self: any, detail_info: DetailInfo, compare_data: dump.CompareData,
                 output_path: str) -> None:
        self.detail_info = detail_info
        self.compare_data = compare_data
        self.detail_writer = DetailWriter(output_path, detail_info)
        self.fusion_op = None

    @staticmethod
    def _check_index_valid(tensor_data, op_name, tensor_index, tensor_type):
        if tensor_index >= len(tensor_data):
            log.print_out_of_range_error(op_name, tensor_type, tensor_index, '[0, %d)' % len(tensor_data))
            raise CompareError(CompareError.MSACCUCMP_INDEX_OUT_OF_BOUNDS_ERROR)

    def compare(self: any) -> int:
        """
        Compare detail by op name
        :return error_code
        """
        tensor_id = self.detail_info.tensor_id.get_tensor_id()
        op_name = self.detail_info.tensor_id.op_name
        tensor_type = self.detail_info.tensor_id.tensor_type
        tensor_index = self.detail_info.tensor_id.index
        log.print_info_log('[%s] Start to compare detail for %s.'
                           % (op_name, tensor_id))

        left_path, left_tensor_data, right_tensor_data = self.get_tensor_data(op_name, tensor_type)

        self._check_index_valid(left_tensor_data, op_name, tensor_index, tensor_type)
        self._check_index_valid(right_tensor_data, op_name, tensor_index, tensor_type)

        my_output_shape = tuple(left_tensor_data[tensor_index].shape)
        ground_truth_shape = tuple(right_tensor_data[tensor_index].shape)
        if my_output_shape != ground_truth_shape:
            log.print_error_log("My Output data shape %s not equal to Ground Truth data shape %s."
                                "Can not compare." % (my_output_shape, ground_truth_shape))
            raise CompareError(CompareError.MSACCUCMP_INVALID_DUMP_DATA_ERROR)

        my_output_array = left_tensor_data[tensor_index].data
        ground_truth_array = right_tensor_data[tensor_index].data

        self.detail_info.set_detail_ops(op_name, op_name)
        self.detail_info.check_and_set_format(utils.convert_shape_to_string(
            left_tensor_data[tensor_index].shape),
            left_tensor_data[tensor_index].tensor_format,
            right_tensor_data[tensor_index].tensor_format
        )

        # delete old result
        try:
            self.detail_writer.delete_old_detail_result_files()
        except (OSError, SystemError, ValueError, TypeError, RuntimeError,
                MemoryError, KeyError, IOError) as error:
            log.print_error_log('Failed to delete the old detail result file. %s' % error)
            raise CompareError(CompareError.MSACCUCMP_DELETE_FILE_ERROR) from error

        dump_file = os.path.basename(left_path)
        self.detail_writer.write(my_output_shape, my_output_array, ground_truth_array, dump_file)
        return CompareError.MSACCUCMP_NONE_ERROR

    def get_tensor_data(self, op_name, tensor_type):
        """
        Get tensor form dump data
        :return left_path
        :return left_tensor_data
        :return right_tensor_data
        """
        try:
            left_path, left_data = self.compare_data.get_left_dump_data(op_name)
        except CompareError as compare_error:
            log.print_error_log('Failed to find %s dump data from -m dump data path.' % op_name)
            raise compare_error
        finally:
            pass
        try:
            right_path, right_data = self.compare_data.get_right_dump_data(op_name)
        except CompareError as compare_error:
            log.print_error_log('Failed to find %s dump data from -g dump data path.' % op_name)
            raise compare_error
        finally:
            pass
        log.print_info_log("My Output data path: " + left_path)
        log.print_info_log("Ground Truth data path: " + right_path)
        if tensor_type == ConstManager.INPUT:
            left_tensor_data = left_data.input_data
            right_tensor_data = right_data.input_data
        else:
            left_tensor_data = left_data.output_data
            right_tensor_data = right_data.output_data
        return left_path, left_tensor_data, right_tensor_data
